# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Handle config file and argument parsing
"""
from collections import deque
from configparser import ConfigParser
from pathlib import Path
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
import configparser
import functools
import itertools
import os

CONFIG_FILENAME = "yoyo.ini"
CONFIG_EDITOR_KEY = "editor"
CONFIG_NEW_MIGRATION_COMMAND_KEY = "post_create_command"

INHERIT = "%inherit"
INCLUDE = "%include"


class CircularReferenceError(configparser.Error):
    """
    %include or %inherit directive has created a circular reference.
    """


class CustomInterpolation(configparser.BasicInterpolation):

    defaults = {}  # type:  Dict[str, str]

    def __init__(self, defaults):
        self.defaults = defaults or {}

    def before_get(self, parser, section, option, value, defaults):
        merged_defaults = self.defaults.copy()
        merged_defaults.update(defaults)
        return super(CustomInterpolation, self).before_get(
            parser, section, option, value, merged_defaults
        )


def get_interpolation_defaults(path: Optional[str] = None):
    parser = configparser.ConfigParser()
    defaults = {
        parser.optionxform(k): v.replace("%", "%%")
        for k, v in os.environ.items()
    }
    if path:
        defaults["here"] = os.path.dirname(os.path.abspath(path))
    return defaults


def get_configparser(defaults=None) -> ConfigParser:
    return ConfigParser(interpolation=CustomInterpolation(defaults))


def update_argparser_defaults(parser, defaults):
    """
    Update an ArgumentParser's defaults.

    Unlike ArgumentParser.set_defaults this will only set defaults for
    arguments the parser has configured.
    """
    known_args = {action.dest for action in parser._actions}
    parser.set_defaults(
        **{k: v for k, v in defaults.items() if k in known_args}
    )


def read_config(src: Optional[str]) -> ConfigParser:
    """
    Read the configuration file at ``path``, or return an empty
    ConfigParse object if ``path`` is ``None``.
    """
    if src is None:
        return get_configparser(get_interpolation_defaults())

    path = _make_path(src)
    config = _read_config(path)
    config_files = {path: config}
    merge_paths = deque([path])
    to_process = [
        ((), path, config)
    ]  # type: List[Tuple[Union[Tuple, Tuple[Path]], Path, ConfigParser]]
    while to_process:
        ancestors, path, config = to_process.pop()
        inherits, includes = find_includes(path, config)
        for p in itertools.chain(inherits, includes):
            if p in ancestors:
                raise CircularReferenceError(
                    "{path!r} contains circular references".format(path=path)
                )
            config = _read_config(p)
            config_files[p] = config
            to_process.append((ancestors + (path,), p, config))

        merge_paths.extendleft(inherits)
        merge_paths.extend(includes)

    merged = merge_configs([config_files[p] for p in merge_paths])
    merged.remove_option("DEFAULT", INCLUDE)
    merged.remove_option("DEFAULT", INHERIT)

    return merged


def _make_path(s: str, basepath: Optional[Path] = None) -> Path:
    """
    Return a fully resolved Path. Raises FileNotFoundError if the path does not
    exist.

    """
    if basepath:
        p = basepath.parent / s
    else:
        p = Path(s)

    # Path.resolve has a backward incompatible change in Python 3.6
    #
    # py35: will raise FileNotFoundError if the path doesn't exist.
    # py36: does not raise FileNotFoundError, unless you pass strict=True
    #
    # Passing strict=True would cause a TypeError in Python 3.5, we
    # have to do an explicit exists() check to support both versions.
    p = p.resolve()
    if not p.exists():
        raise FileNotFoundError(p)
    return p.resolve()


def _read_config(path: Path) -> ConfigParser:
    config = get_configparser(get_interpolation_defaults(str(path)))
    config.read([str(path)])
    return config


def find_includes(
    basepath: Path, config: ConfigParser
) -> Tuple[List[Path], List[Path]]:

    result = {INCLUDE: [], INHERIT: []}  # type: Dict[str, List[Path]]
    for key in [INHERIT, INCLUDE]:
        try:
            paths = config["DEFAULT"][key].split()
        except KeyError:
            continue

        for p in paths:
            strict = True
            if p.startswith("?"):
                strict = False
                p = p[1:]
            try:
                path = _make_path(p, basepath)
            except FileNotFoundError:
                if strict:
                    raise
                continue

            result[key].append(path)
    return result[INHERIT], result[INCLUDE]


def merge_configs(configs: List[ConfigParser],) -> ConfigParser:
    def merge(c1, c2):
        c1.read_dict(c2)
        return c1

    return functools.reduce(merge, configs[1:], configs[0])


def save_config(config, path):
    """
    Write the configuration file to ``path``.
    """
    os.umask(0o77)
    f = open(path, "w")
    try:
        return config.write(f)
    finally:
        f.close()


def find_config():
    """Find the closest config file in the cwd or a parent directory"""
    d = os.getcwd()
    while d != os.path.dirname(d):
        path = os.path.join(d, CONFIG_FILENAME)
        if os.path.isfile(path):
            return path
        d = os.path.dirname(d)
    return None



from datetime import datetime
import getpass
import socket

from yoyo import internalmigrations
from yoyo.tests import clear_database


def assert_table_is_created(backend, table):
    assert table in backend.list_tables()


def assert_table_is_missing(backend, table):
    assert table not in backend.list_tables()


def test_it_installs_migrations_table(backend):
    clear_database(backend)
    internalmigrations.upgrade(backend)


def test_it_installs_v1(backend):
    clear_database(backend)
    internalmigrations.upgrade(backend, version=1)
    assert internalmigrations.get_current_version(backend) == 1
    assert_table_is_created(backend, "_yoyo_migration")
    assert_table_is_missing(backend, "_yoyo_version")
    assert_table_is_missing(backend, "_yoyo_log")


def test_it_installs_v2(backend):
    clear_database(backend)
    internalmigrations.upgrade(backend, version=2)
    assert internalmigrations.get_current_version(backend) == 2
    assert_table_is_created(backend, "_yoyo_migration")
    assert_table_is_created(backend, "_yoyo_version")
    assert_table_is_created(backend, "_yoyo_log")


def test_v3_preserves_history_when_upgrading(backend):
    clear_database(backend)
    internalmigrations.upgrade(backend, version=1)
    v2insert = (
        "INSERT INTO {0.migration_table_quoted} (id, ctime) "
        "VALUES (:id, :when)".format(backend)
    )

    with backend.transaction():
        backend.execute(
            v2insert, {"id": "migration-a", "when": datetime(2000, 1, 1, 12)}
        )
        backend.execute(
            v2insert, {"id": "migration-b", "when": datetime(2000, 2, 1, 12)}
        )

    internalmigrations.upgrade(backend, version=2)

    cursor = backend.execute(
        "SELECT migration_hash, migration_id, applied_at_utc "
        "FROM {0.migration_table_quoted} order by applied_at_utc".format(
            backend
        )
    )
    applied = list(cursor.fetchall())
    assert applied == [
        (
            "da335f4748bac940534a07dcd997122ff9fc7c7ebe90b78f5c11aeda2a2d0104",
            "migration-a",
            datetime(2000, 1, 1, 12),
        ),
        (
            "82f4683f1cc4f376e7c5a996392927d9a5ed6d102681cd8609b65b621e72ed78",
            "migration-b",
            datetime(2000, 2, 1, 12),
        ),
    ]

    cursor = backend.execute(
        "SELECT migration_hash, migration_id, "
        "operation, created_at_utc, username, hostname, comment "
        "FROM {0.log_table_quoted} order by created_at_utc".format(backend)
    )
    log = list(cursor.fetchall())
    current_user = getpass.getuser()
    current_host = socket.getfqdn()
    assert log == [
        (
            "da335f4748bac940534a07dcd997122ff9fc7c7ebe90b78f5c11aeda2a2d0104",
            "migration-a",
            "apply",
            datetime(2000, 1, 1, 12),
            current_user,
            current_host,
            "this log entry created automatically by an internal schema upgrade",
        ),
        (
            "82f4683f1cc4f376e7c5a996392927d9a5ed6d102681cd8609b65b621e72ed78",
            "migration-b",
            "apply",
            datetime(2000, 2, 1, 12),
            current_user,
            current_host,
            "this log entry created automatically by an internal schema upgrade",
        ),
    ]



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "ancestors",
    "default_migration_table",
    "descendants",
    "get_backend",
    "group",
    "logger",
    "read_migrations",
    "step",
    "transaction",
]

from yoyo.connections import get_backend
from yoyo.migrations import ancestors
from yoyo.migrations import default_migration_table
from yoyo.migrations import descendants
from yoyo.migrations import group
from yoyo.migrations import logger
from yoyo.migrations import read_migrations
from yoyo.migrations import step
from yoyo.migrations import transaction

__version__ = "7.3.1"



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
from urllib.parse import parse_qsl
from urllib.parse import quote
from urllib.parse import unquote
from urllib.parse import urlencode
from urllib.parse import urlsplit
from urllib.parse import urlunsplit

from .migrations import default_migration_table
from .backends import PostgresqlBackend
from .backends import RedshiftBackend
from .backends import SQLiteBackend
from .backends import ODBCBackend
from .backends import OracleBackend
from .backends import MySQLBackend
from .backends import MySQLdbBackend
from .backends import SnowflakeBackend

BACKENDS = {
    "odbc": ODBCBackend,
    "oracle": OracleBackend,
    "postgresql": PostgresqlBackend,
    "postgres": PostgresqlBackend,
    "psql": PostgresqlBackend,
    "mysql": MySQLBackend,
    "mysql+mysqldb": MySQLdbBackend,
    "sqlite": SQLiteBackend,
    "snowflake": SnowflakeBackend,
    "redshift": RedshiftBackend,
}


_DatabaseURI = namedtuple(
    "_DatabaseURI", "scheme username password hostname port database args"
)


class DatabaseURI(_DatabaseURI):
    @property
    def netloc(self):
        hostname = self.hostname or ""
        if self.port:
            hostpart = "{}:{}".format(hostname, self.port)
        else:
            hostpart = hostname

        if self.username:
            return "{}:{}@{}".format(
                quote(self.username, safe=""),
                quote(self.password or "", safe=""),
                hostpart,
            )
        else:
            return hostpart

    def __str__(self):
        return urlunsplit(
            (self.scheme, self.netloc, self.database, urlencode(self.args), "")
        )

    @property
    def uri(self):
        return str(self)


class BadConnectionURI(Exception):
    """
    An invalid connection URI
    """


def get_backend(uri, migration_table=default_migration_table):
    """
    Connect to the given DB uri in the format
    ``driver://user:pass@host:port/database_name?param=value``,
    returning a :class:`DatabaseBackend` object
    """
    parsed = parse_uri(uri)
    try:
        backend_class = BACKENDS[parsed.scheme.lower()]
    except KeyError:
        raise BadConnectionURI(
            "Unrecognised database connection scheme %r" % parsed.scheme
        )
    return backend_class(parsed, migration_table)


def parse_uri(s):
    """
    Examples::

        >>> parse_uri('postgres://fred:bassett@server:5432/fredsdatabase')
        ('postgres', 'fred', 'bassett', 'server', 5432, 'fredsdatabase', None)
        >>> parse_uri('mysql:///jimsdatabase')
        ('mysql', None, None, None, None, 'jimsdatabase', None, None)
        >>> parse_uri('odbc://user:password@server/database?DSN=dsn')
        ('odbc', 'user', 'password', 'server', None, 'database', {'DSN':'dsn'})
    """
    result = urlsplit(s)

    if not result.scheme:
        raise BadConnectionURI("No scheme specified in connection URI %r" % s)

    return DatabaseURI(
        scheme=result.scheme,
        username=(
            unquote(result.username) if result.username is not None else None
        ),
        password=(
            unquote(result.password) if result.password is not None else None
        ),
        hostname=result.hostname,
        port=result.port,
        database=result.path[1:] if result.path else None,
        args=dict(parse_qsl(result.query)),
    )



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from shutil import rmtree
from datetime import datetime
from tempfile import mkdtemp
from functools import partial
from itertools import count
import io
import os
import os.path
import re
import sys
import textwrap

from mock import Mock, patch, call
import frozendate
import pytest
import tms

from yoyo import read_migrations
from yoyo.config import get_configparser
from yoyo.tests import with_migrations, dburi
from yoyo.tests import migrations_dir
from yoyo.tests import get_backend
from yoyo.scripts.main import main, parse_args, LEGACY_CONFIG_FILENAME
from yoyo.scripts import newmigration


def is_tmpfile(p, directory=None):
    return (
        p.startswith(directory) if directory else True
    ) and os.path.basename(p).startswith(newmigration.tempfile_prefix)


class TestInteractiveScript(object):
    def setup(self):
        self.stdout_tty_patch = patch("sys.stdout.isatty", return_value=True)
        self.stdout_tty_patch.start()
        self.confirm_patch = patch("yoyo.utils.confirm", return_value=False)
        self.confirm = self.confirm_patch.start()
        self.prompt_patch = patch("yoyo.utils.prompt", return_value="n")
        self.prompt = self.prompt_patch.start()
        self.tmpdir = mkdtemp()
        self.dburi = "sqlite:////" + self.tmpdir + "db.sqlite"
        self.saved_cwd = os.getcwd()
        os.chdir(self.tmpdir)

    def teardown(self):
        self.prompt_patch.stop()
        self.confirm_patch.stop()
        self.stdout_tty_patch.stop()
        os.chdir(self.saved_cwd)
        rmtree(self.tmpdir)

    def writeconfig(self, **defaults):
        cp = get_configparser()
        for item in defaults:
            cp.set("DEFAULT", item, defaults[item])

        if sys.version_info < (3, 0):
            with open("yoyo.ini", "w") as f:
                cp.write(f)
        else:
            with open("yoyo.ini", "w", encoding="UTF-8") as f:
                cp.write(f)

    def get_migration_log(self):
        return (
            get_backend(self.dburi)
            .execute("SELECT migration_id, operation FROM _yoyo_log")
            .fetchall()
        )


class TestYoyoScript(TestInteractiveScript):
    @with_migrations()
    def test_it_sets_verbosity_level(self, tmpdir):
        with patch("yoyo.scripts.main.configure_logging") as m:
            main(["apply", tmpdir, "--database", dburi])
            assert m.call_args == call(0)
            main(["-vvv", "apply", tmpdir, "--database", dburi])
            assert m.call_args == call(3)

    @with_migrations()
    def test_it_prompts_to_create_config_file(self, tmpdir):
        main(["apply", tmpdir, "--database", dburi])
        assert "save migration config" in self.confirm.call_args[0][0].lower()

    @with_migrations()
    def test_it_creates_config_file(self, tmpdir):
        self.confirm.return_value = True
        main(["apply", tmpdir, "--database", dburi])
        assert os.path.exists("yoyo.ini")
        with open("yoyo.ini") as f:
            assert "database = {0}".format(dburi) in f.read()

    @with_migrations()
    def test_it_uses_config_file(self, tmpdir):
        self.writeconfig(batch_mode="on")
        with patch("yoyo.scripts.migrate.apply") as apply:
            main(["apply", tmpdir, "--database", dburi])
            args_used = apply.call_args[0][0]
            assert args_used.batch_mode is True

    @with_migrations()
    def test_it_ignores_config_file(self, tmpdir):
        self.writeconfig(batch_mode="on")
        with patch("yoyo.scripts.migrate.apply") as apply:
            main(["apply", "--no-config-file", tmpdir, "--database", dburi])
            args_used = apply.call_args[0][0]
            assert args_used.batch_mode is False

    @with_migrations()
    def test_it_prompts_password(self, tmpdir):
        dburi = "sqlite://user@/:memory"
        with patch(
            "yoyo.scripts.main.getpass", return_value="fish"
        ) as getpass, patch("yoyo.connections.get_backend") as get_backend:
            main(["apply", tmpdir, "--database", dburi, "--prompt-password"])
            assert getpass.call_count == 1
            assert get_backend.call_args == call(
                "sqlite://user:fish@/:memory", "_yoyo_migration"
            )

    @with_migrations()
    def test_it_prompts_migrations(self, tmpdir):
        with patch(
            "yoyo.scripts.migrate.prompt_migrations"
        ) as prompt_migrations, patch(
            "yoyo.scripts.migrate.get_backend"
        ) as get_backend:
            main(["apply", tmpdir, "--database", dburi])
            migrations = get_backend().to_apply()
            assert migrations in prompt_migrations.call_args[0]

    @with_migrations()
    def test_it_applies_migrations(self, tmpdir):
        with patch("yoyo.scripts.migrate.get_backend") as get_backend:
            main(["-b", "apply", tmpdir, "--database", dburi])
            assert get_backend().rollback_migrations.call_count == 0
            assert get_backend().apply_migrations.call_count == 1

    @with_migrations()
    def test_it_rollsback_migrations(self, tmpdir):
        with patch("yoyo.scripts.migrate.get_backend") as get_backend:
            main(["-b", "rollback", tmpdir, "--database", dburi])
            assert get_backend().rollback_migrations.call_count == 1
            assert get_backend().apply_migrations.call_count == 0

    @with_migrations()
    def test_it_reapplies_migrations(self, tmpdir):
        with patch("yoyo.scripts.migrate.get_backend") as get_backend:
            main(["-b", "reapply", tmpdir, "--database", dburi])
            assert get_backend().rollback_migrations.call_count == 1
            assert get_backend().apply_migrations.call_count == 1

    @with_migrations(m1='step("CREATE TABLE yoyo_test1 (id INT)")')
    @with_migrations(m2='step("CREATE TABLE yoyo_test2 (id INT)")')
    def test_it_applies_from_multiple_sources(self, t1, t2):
        with patch("yoyo.backends.DatabaseBackend.apply_migrations") as apply:
            main(["-b", "apply", t1, t2, "--database", dburi])
            call_posargs, call_kwargs = apply.call_args
            migrations, _ = call_posargs
            assert [m.path for m in migrations] == [
                os.path.join(t1, "m1.py"),
                os.path.join(t2, "m2.py"),
            ]

    @with_migrations()
    def test_it_offers_to_upgrade(self, tmpdir):
        legacy_config_path = os.path.join(tmpdir, LEGACY_CONFIG_FILENAME)
        with io.open(legacy_config_path, "w", encoding="utf-8") as f:
            f.write("[DEFAULT]\n")
            f.write("migration_table=_yoyo_migration\n")
            f.write("dburi=sqlite:///\n")

        self.confirm.return_value = True
        main(["apply", tmpdir])
        prompts = [
            args[0].lower() for args, kwargs in self.confirm.call_args_list
        ]
        assert len(prompts) == 2
        assert prompts[0].startswith("move legacy configuration")
        assert prompts[1].startswith("delete legacy configuration")
        assert not os.path.exists(legacy_config_path)

        with open("yoyo.ini", "r") as f:
            config = f.read()
            assert "database = sqlite:///\n" in config
            assert "migration_table = _yoyo_migration\n" in config
            assert "batch_mode = off\n" in config
            assert "verbosity = 0\n" in config

    @with_migrations()
    def test_it_upgrades_migration_table_None(self, tmpdir):
        legacy_config_path = os.path.join(tmpdir, LEGACY_CONFIG_FILENAME)
        with io.open(legacy_config_path, "w", encoding="utf-8") as f:
            f.write("[DEFAULT]\n")
            f.write("migration_table=None\n")
            f.write("dburi=sqlite:///\n")
        self.confirm.return_value = True
        main(["apply", tmpdir])

        with open("yoyo.ini", "r") as f:
            config = f.read()
        assert "migration_table = _yoyo_migration\n" in config

    @with_migrations()
    def test_it_forces_batch_mode_if_not_running_in_a_tty(self, tmpdir):
        with patch("sys.stdout", isatty=lambda: False):
            main(["apply", tmpdir, "--database", dburi])
            assert self.prompt.call_count == 0
            assert self.confirm.call_count == 0

    def test_concurrent_instances_do_not_conflict(self, backend):
        import threading
        from functools import partial

        if backend.uri.scheme == "sqlite":
            pytest.skip("Concurrency tests not supported for sqlite databases")

        with migrations_dir(
            m1=(
                "import time\n"
                "step(lambda conn: time.sleep(0.1))\n"
                "step(\"INSERT INTO yoyo_t VALUES ('A')\")"
            )
        ) as tmpdir:
            assert "yoyo_t" in backend.list_tables()
            backend.rollback()
            backend.execute("SELECT * FROM yoyo_t")
            run_migrations = partial(
                main, ["apply", "-b", tmpdir, "--database", str(backend.uri)]
            )
            threads = [
                threading.Thread(target=run_migrations) for ix in range(20)
            ]
            for t in threads:
                t.start()
            for t in threads:
                t.join()

            # Exactly one instance of the migration script should have succeeded
            backend.rollback()
            cursor = backend.execute("SELECT COUNT(1) from yoyo_t")
            assert cursor.fetchone()[0] == 1

    def test_it_breaks_lock(self, dburi):
        if dburi.startswith("sqlite"):
            pytest.skip("Test not supported for sqlite databases")
        backend = get_backend(dburi)
        backend.execute(
            "INSERT INTO yoyo_lock (locked, ctime, pid) " "VALUES (1, :now, 1)",
            {"now": datetime.utcnow()},
        )
        backend.commit()
        main(["break-lock", "--database", dburi])
        lock_count = backend.execute(
            "SELECT COUNT(1) FROM yoyo_lock"
        ).fetchone()[0]
        assert lock_count == 0

    def test_it_prompts_password_on_break_lock(self, tmpdir):
        dburi = "sqlite://user@/:memory"
        with patch(
            "yoyo.scripts.main.getpass", return_value="fish"
        ) as getpass, patch("yoyo.connections.get_backend") as get_backend:
            main(["break-lock", "--database", dburi, "--prompt-password"])
            assert getpass.call_count == 1
            assert get_backend.call_args == call(
                "sqlite://user:fish@/:memory", "_yoyo_migration"
            )


class TestArgParsing(TestInteractiveScript):
    def test_it_uses_config_file_defaults(self):
        self.writeconfig(
            sources="/tmp/migrations",
            database="postgresql:///foo",
            migration_table="my_migrations",
        )
        _, _, args = parse_args(["apply"])
        assert args.database == "postgresql:///foo"
        assert args.sources == ["/tmp/migrations"]
        assert args.migration_table == "my_migrations"

    def test_it_uses_interpolated_values_from_config(self):
        self.writeconfig(sources="%(here)s/migrations")
        _, _, args = parse_args(["apply"])
        assert args.sources == [os.getcwd() + "/migrations"]

    def test_cli_args_take_precendence(self):
        self.writeconfig(sources="A")
        _, _, args = parse_args(["apply", "B", "--database", "C"])
        assert args.sources == ["B"]

    def test_global_args_can_appear_before_command(self):
        _, _, args = parse_args(["apply", "X", "--database", "Y"])
        assert args.verbosity == 0
        _, _, args = parse_args(["-v", "apply", "X", "--database", "Y"])
        assert args.verbosity == 1

    def test_global_args_can_appear_after_command(self):
        _, _, args = parse_args(["apply", "X", "Y"])
        assert args.verbosity == 0
        _, _, args = parse_args(["apply", "-v", "X", "Y"])
        assert args.verbosity == 1


class TestMarkCommand(TestInteractiveScript):
    @with_migrations(
        m1='step("CREATE TABLE test1 (id INT)")',
        m2='step("CREATE TABLE test2 (id INT)")',
        m3='step("CREATE TABLE test3 (id INT)")',
    )
    def test_it_prompts_only_unapplied(self, tmpdir):
        from yoyo.connections import get_backend

        migrations = read_migrations(tmpdir)
        backend = get_backend(self.dburi)
        backend.apply_migrations(migrations[:1])

        with patch(
            "yoyo.scripts.migrate.prompt_migrations"
        ) as prompt_migrations:
            main(["mark", tmpdir, "--database", self.dburi])
            _, prompted, _ = prompt_migrations.call_args[0]
            prompted = [m.id for m in prompted]
            assert prompted == ["m2", "m3"]

    @with_migrations(
        m1='step("INSERT INTO t VALUES (1)")',
        m2='__depends__=["m1"]; step("INSERT INTO t VALUES (2)")',
        m3='step("INSERT INTO t VALUES (2)")',
    )
    def test_it_marks_at_selected_version(self, tmpdir):
        from yoyo.connections import get_backend

        self.confirm.return_value = True
        migrations = read_migrations(tmpdir)
        backend = get_backend(self.dburi)
        with backend.transaction():
            backend.execute("CREATE TABLE t (id INT)")

        main(["mark", "-r", "m2", tmpdir, "--database", self.dburi])
        assert backend.is_applied(migrations[0])
        assert backend.is_applied(migrations[1])
        assert not backend.is_applied(migrations[2])

        # Check that migration steps have not been applied
        c = backend.execute("SELECT * FROM t")
        assert len(c.fetchall()) == 0


class TestUnmarkCommand(TestInteractiveScript):
    @with_migrations(m1="", m2="", m3="")
    def test_it_prompts_only_applied(self, tmpdir):
        from yoyo.connections import get_backend

        migrations = read_migrations(tmpdir)
        backend = get_backend(self.dburi)
        backend.apply_migrations(migrations[:2])
        assert len(backend.get_applied_migration_hashes()) == 2

        with patch(
            "yoyo.scripts.migrate.prompt_migrations"
        ) as prompt_migrations:
            main(["unmark", tmpdir, "--database", self.dburi])
            _, prompted, _ = prompt_migrations.call_args[0]
            prompted = [m.id for m in prompted]
            assert prompted == ["m2", "m1"]

    @with_migrations(m1="", m2='__depends__=["m1"]', m3='__depends__=["m2"]')
    def test_it_unmarks_to_selected_revision(self, tmpdir):
        from yoyo.connections import get_backend

        self.confirm.return_value = True
        migrations = read_migrations(tmpdir)
        backend = get_backend(self.dburi)
        backend.apply_migrations(migrations)

        main(["unmark", "-r", "m2", tmpdir, "--database", self.dburi])
        assert backend.is_applied(migrations[0])
        assert not backend.is_applied(migrations[1])
        assert not backend.is_applied(migrations[2])


class TestNewMigration(TestInteractiveScript):
    def setup(self):
        def mockstat(f, c=count()):
            return Mock(st_mtime=next(c))

        super(TestNewMigration, self).setup()
        self.subprocess_patch = patch("yoyo.scripts.newmigration.subprocess")
        self.subprocess = self.subprocess_patch.start()
        self.subprocess.call.return_value = 0
        self.stat_patch = patch("yoyo.scripts.newmigration.stat", mockstat)
        self.stat_patch.start()

    def teardown(self):
        super(TestNewMigration, self).teardown()
        self.subprocess_patch.stop()
        self.stat_patch.stop()

    @with_migrations()
    def test_it_creates_an_empty_migration(self, tmpdir):
        main(["new", "-b", "-m", "foo", tmpdir, "--database", dburi])
        assert any("-foo.py" in f for f in os.listdir(tmpdir))

    @with_migrations(
        m1="", m2='__depends__=["m1"]; step("INSERT INTO t VALUES (2)")', m3=""
    )
    def test_it_depends_on_all_current_heads(self, tmpdir):
        main(["new", "-b", "-m", "foo", tmpdir, "--database", dburi])
        m = next(f for f in os.listdir(tmpdir) if "-foo.py" in f)
        with io.open(os.path.join(tmpdir, m), encoding="utf-8") as f:
            assert "__depends__ = {'m2', 'm3'}" in f.read()

    @with_migrations()
    def test_it_names_file_by_date_and_sequence(self, tmpdir):
        with frozendate.freeze(2001, 1, 1):
            main(["new", "-b", "-m", "foo", tmpdir, "--database", dburi])
            main(["new", "-b", "-m", "bar", tmpdir, "--database", dburi])
        names = [n for n in sorted(os.listdir(tmpdir)) if n.endswith(".py")]
        assert names[0].startswith("20010101_01_")
        assert names[0].endswith("-foo.py")
        assert names[1].startswith("20010101_02_")
        assert names[1].endswith("-bar.py")

    @with_migrations()
    def test_it_invokes_correct_editor_binary_from_config(self, tmpdir):
        self.writeconfig(editor="vim {} -c +10")
        main(["new", tmpdir, "--database", dburi])
        assert self.subprocess.call.call_args == call(
            [
                "vim",
                tms.Matcher(partial(is_tmpfile, directory=tmpdir)),
                "-c",
                "+10",
            ]
        )

    @with_migrations()
    def test_it_invokes_correct_editor_binary_from_env(self, tmpdir):
        # default to $VISUAL
        with patch("os.environ", {"EDITOR": "ed", "VISUAL": "visualed"}):
            main(["new", tmpdir, "--database", dburi])
            assert self.subprocess.call.call_args == call(
                ["visualed", tms.Unicode()]
            )

        # fallback to $EDITOR
        with patch("os.environ", {"EDITOR": "ed"}):
            main(["new", tmpdir, "--database", dburi])
            assert self.subprocess.call.call_args == call(["ed", tms.Unicode()])

        # Otherwise, vi
        with patch("os.environ", {}):
            main(["new", tmpdir, "--database", dburi]) == call(
                ["vi", tms.Unicode()]
            )

        # Prompts should only appear if there is an error reading the migration
        # file, which should not be the case.
        assert self.prompt.call_args_list == []

    @with_migrations()
    def test_it_pulls_message_from_docstring(self, tmpdir):
        def write_migration(argv):
            with io.open(argv[-1], "w", encoding="utf8") as f:
                f.write('"""\ntest docstring\nsplit over\n\nlines\n"""\n')

        self.subprocess.call = write_migration
        main(["new", tmpdir, "--database", dburi])
        names = [n for n in sorted(os.listdir(tmpdir)) if n.endswith(".py")]
        assert "test-docstring" in names[0]

    @with_migrations()
    def test_it_prompts_to_reedit_bad_migration(self, tmpdir):
        def write_migration(argv):
            with io.open(argv[-1], "w", encoding="utf8") as f:
                f.write("this is not valid python!")

        self.subprocess.call = write_migration
        main(["new", tmpdir, "--database", dburi])
        prompts = [
            args[0].lower() for args, kwargs in self.prompt.call_args_list
        ]
        assert "retry editing?" in prompts[0]

    @with_migrations()
    def test_it_defaults_docstring_to_message(self, tmpdir):
        main(["new", "-b", "-m", "your ad here", tmpdir, "--database", dburi])
        names = [n for n in sorted(os.listdir(tmpdir)) if n.endswith(".py")]
        with io.open(
            os.path.join(tmpdir, names[0]), "r", encoding="utf-8"
        ) as f:
            assert "your ad here" in f.read()

    @with_migrations()
    def test_it_calls_post_create_command(self, tmpdir):
        self.writeconfig(post_create_command="/bin/ls -l {} {}")
        with frozendate.freeze(2001, 1, 1):
            main(["new", "-b", tmpdir, "--database", dburi])
        is_filename = tms.Str(
            lambda s: os.path.basename(s).startswith("20010101_01_")
        )
        assert self.subprocess.call.call_args == call(
            ["/bin/ls", "-l", is_filename, is_filename]
        )

    @with_migrations()
    def test_it_uses_configured_prefix(self, tmpdir):
        self.writeconfig(prefix="foo_")
        main(["new", "-b", "-m", "bar", tmpdir, "--database", dburi])
        names = [n for n in sorted(os.listdir(tmpdir)) if n.endswith(".py")]
        assert re.match("foo_.*-bar", names[0]) is not None

    @with_migrations(m1="")
    def test_it_creates_sql_file(self, tmpdir):
        main(
            [
                "new",
                "-b",
                "-m",
                "comment",
                "--sql",
                tmpdir,
                "--database",
                dburi,
            ]
        )
        name = next(n for n in sorted(os.listdir(tmpdir)) if n.endswith(".sql"))
        with io.open(os.path.join(tmpdir, name), "r", encoding="utf-8") as f:
            assert f.read() == textwrap.dedent(
                """\
                -- comment
                -- depends: m1

                """
            )


class TestList(TestInteractiveScript):
    def get_output(self, migrations):
        with migrations_dir(**migrations) as tmpdir:
            with patch("sys.stdout", io.StringIO()) as captured:
                main(["list", tmpdir, "--database", dburi])
            return captured.getvalue()

    def test_it_lists_migrations(self, tmpdir):
        output = self.get_output({"m1": "", "m2": ""})
        assert re.search(r"^U\s+m1", output, re.M)
        assert re.search(r"^U\s+m2", output, re.M)


class TestDevelopCommand(TestInteractiveScript):
    def test_it_applies_outstanding_migrations(self):
        with migrations_dir(m1="", m2="") as tmpdir:
            main(["develop", tmpdir, "--database", self.dburi])
            assert self.get_migration_log() == [
                ("m1", "apply"),
                ("m2", "apply"),
            ]

    def test_it_reapplies_last_migration(self, tmpdir):
        with migrations_dir(
            m1='step("CREATE TABLE yoyo_test1 (id INT)", "DROP TABLE yoyo_test1")',
            m2='step("CREATE TABLE yoyo_test2 (id INT)", "DROP TABLE yoyo_test2")',
        ) as tmpdir:
            main(["-b", "apply", tmpdir, "--database", self.dburi])
            assert self.get_migration_log() == [
                ("m1", "apply"),
                ("m2", "apply"),
            ]
            main(["develop", tmpdir, "--database", self.dburi])
            assert self.get_migration_log() == [
                ("m1", "apply"),
                ("m2", "apply"),
                ("m2", "rollback"),
                ("m2", "apply"),
            ]

    def test_it_reapplies_last_n_migrations(self, tmpdir):
        with migrations_dir(
            m1='step("CREATE TABLE yoyo_test1 (id INT)", "DROP TABLE yoyo_test1")',
            m2='step("CREATE TABLE yoyo_test2 (id INT)", "DROP TABLE yoyo_test2")',
        ) as tmpdir:
            main(["-b", "apply", tmpdir, "--database", self.dburi])
            assert self.get_migration_log() == [
                ("m1", "apply"),
                ("m2", "apply"),
            ]
            main(["develop", tmpdir, "-n", "2", "--database", self.dburi])
            assert self.get_migration_log() == [
                ("m1", "apply"),
                ("m2", "apply"),
                ("m2", "rollback"),
                ("m1", "rollback"),
                ("m1", "apply"),
                ("m2", "apply"),
            ]



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from getpass import getpass
import argparse
import configparser
import logging
import os
import sys

from yoyo import connections
from yoyo import default_migration_table
from yoyo import logger
from yoyo import utils
from yoyo.config import CONFIG_FILENAME
from yoyo.config import find_config
from yoyo.config import read_config
from yoyo.config import save_config
from yoyo.config import update_argparser_defaults

verbosity_levels = {
    0: logging.ERROR,
    1: logging.WARN,
    2: logging.INFO,
    3: logging.DEBUG,
}

min_verbosity = min(verbosity_levels)
max_verbosity = max(verbosity_levels)

LEGACY_CONFIG_FILENAME = ".yoyo-migrate"


class InvalidArgument(Exception):
    pass


def parse_args(argv=None):
    """
    Parse the config file and command line args.

    :return: tuple of (argparser, parsed_args)
    """
    #: List of arguments whose defaults should be read from the config file
    config_args = {
        "batch_mode": "getboolean",
        "sources": "get",
        "database": "get",
        "verbosity": "getint",
        "migration_table": "get",
    }

    globalparser, argparser, subparsers = make_argparser()

    # Initial parse to extract --config and any global arguments
    global_args, _ = globalparser.parse_known_args(argv)

    # Read the config file and create a dictionary of defaults for argparser
    config = read_config(
        (global_args.config or find_config())
        if global_args.use_config_file
        else None
    )

    defaults = {}
    for argname, getter in config_args.items():
        try:
            defaults[argname] = getattr(config, getter)("DEFAULT", argname)
        except configparser.NoOptionError:
            pass

    if "sources" in defaults:
        defaults["sources"] = defaults["sources"].split()

    # Set the argparser defaults to values read from the config file
    update_argparser_defaults(globalparser, defaults)
    update_argparser_defaults(argparser, defaults)
    for subp in subparsers.choices.values():
        update_argparser_defaults(subp, defaults)

    # Now parse for real, starting from the top
    args = argparser.parse_args(argv)

    # Update the args namespace with the global args.
    # This ensures that global args (eg '-v) are recognized regardless
    # of whether they were placed before or after the subparser command.
    # If we do not do this then the sub_parser copy of the argument takes
    # precedence, and overwrites any global args set before the command name.
    args.__dict__.update(globalparser.parse_known_args(argv)[0].__dict__)

    return config, argparser, args


def make_argparser():
    """
    Return a top-level ArgumentParser parser object,
    plus a list of sub_parsers
    """
    global_parser = argparse.ArgumentParser(add_help=False)
    global_parser.add_argument(
        "--config", "-c", default=None, help="Path to config file"
    )
    global_parser.add_argument(
        "-v",
        dest="verbosity",
        action="count",
        default=min_verbosity,
        help="Verbose output. Use multiple times "
        "to increase level of verbosity",
    )
    global_parser.add_argument(
        "-b",
        "--batch",
        dest="batch_mode",
        action="store_true",
        default=(not sys.stdout.isatty()),
        help="Run in batch mode" ". Turns off all user prompts",
    )

    global_parser.add_argument(
        "--no-config-file",
        "--no-cache",
        dest="use_config_file",
        action="store_false",
        default=True,
        help="Don't look for a yoyo.ini config file",
    )
    argparser = argparse.ArgumentParser(prog="yoyo", parents=[global_parser])

    subparsers = argparser.add_subparsers(help="Commands help")

    from . import migrate, newmigration

    migrate.install_argparsers(global_parser, subparsers)
    newmigration.install_argparsers(global_parser, subparsers)

    return global_parser, argparser, subparsers


def configure_logging(level):
    """
    Configure the python logging module with the requested loglevel
    """
    logging.basicConfig(level=verbosity_levels[level])


def prompt_save_config(config, path):
    # Offer to save the current configuration for future runs
    # Don't cache anything in batch mode (because we can't prompt to find the
    # user's preference).

    if utils.confirm(
        "Save migration configuration to {}?\n"
        "This is saved in plain text and "
        "contains your database password.\n\n"
        "Answering 'y' means you do not have to specify "
        "the migration source or database connection "
        "for future runs".format(path)
    ):
        save_config(config, path)


def upgrade_legacy_config(args, config, sources):

    for dir in reversed(sources):
        path = os.path.join(dir, LEGACY_CONFIG_FILENAME)
        if not os.path.isfile(path):
            continue

        legacy_config = read_config(path)

        def transfer_setting(
            oldname, newname, transform=None, section="DEFAULT"
        ):
            try:
                config.get(section, newname)
            except configparser.NoOptionError:
                try:
                    value = legacy_config.get(section, oldname)
                except configparser.NoOptionError:
                    pass
                else:
                    if transform:
                        value = transform(value)
                    config.set(section, newname, value)

        transfer_setting("dburi", "database")
        transfer_setting(
            "migration_table",
            "migration_table",
            lambda v: (default_migration_table if v == "None" else v),
        )

        config_path = args.config or CONFIG_FILENAME
        if not args.batch_mode:
            if utils.confirm(
                "Move legacy configuration in {!r} to {!r}?".format(
                    path, config_path
                )
            ):
                save_config(config, config_path)
                try:
                    if utils.confirm(
                        "Delete legacy configuration file {!r}".format(path)
                    ):
                        os.unlink(path)
                except OSError:
                    logger.warn(
                        "Could not remove %r. Manually remove this file "
                        "to avoid future warnings",
                        path,
                    )
                return True
        else:
            logger.warn(
                "Found legacy configuration in %r. Run "
                "yoyo in interactive mode to update your "
                "configuration files",
                path,
            )

            try:
                args.database = args.database or legacy_config.get(
                    "DEFAULT", "dburi"
                )
            except configparser.NoOptionError:
                pass
            try:
                if not vars(args).get("migration_table"):
                    args.migration_table = legacy_config.get(
                        "DEFAULT", "migration_table"
                    )
            except configparser.NoOptionError:
                pass

        return False


def get_backend(args, config):
    try:
        dburi = args.database
    except AttributeError:
        dburi = config.get("DEFAULT", "database")

    try:
        migration_table = args.migration_table
    except AttributeError:
        try:
            migration_table = config.get("DEFAULT", "migration_table")
        except configparser.NoOptionError:
            migration_table = default_migration_table

    if dburi is None:
        raise InvalidArgument("Please specify a database uri")

    try:
        if args.prompt_password:
            password = getpass("Password for %s: " % dburi)
            parsed = connections.parse_uri(dburi)
            dburi = parsed._replace(password=password).uri
    except AttributeError:
        pass

    return connections.get_backend(dburi, migration_table)


def main(argv=None):
    config, argparser, args = parse_args(argv)

    if getattr(args, "func", None) is None:
        argparser.print_usage(sys.stderr)
        argparser.exit(1)

    config_is_empty = config.sections() == [] and config.items("DEFAULT") == []

    sources = getattr(args, "sources", None)

    verbosity = args.verbosity
    verbosity = min(max_verbosity, max(min_verbosity, verbosity))
    configure_logging(verbosity)

    if vars(args).get("sources"):
        config.set("DEFAULT", "sources", " ".join(args.sources))
    if vars(args).get("database"):
        # ConfigParser requires that any percent signs in the db uri be escaped.
        config.set("DEFAULT", "database", args.database.replace("%", "%%"))
    if vars(args).get("migration_table"):
        config.set("DEFAULT", "migration_table", args.migration_table)
    config.set(
        "DEFAULT",
        "batch_mode",
        "on" if vars(args).get("batch_mode") else "off",
    )
    config.set("DEFAULT", "verbosity", str(vars(args).get("verbosity")))

    if sources:
        if upgrade_legacy_config(args, config, sources):
            return main(argv)

    try:
        if vars(args).get("func"):
            args.func(args, config)
    except InvalidArgument as e:
        argparser.error(e.args[0])

    if config_is_empty and args.use_config_file and not args.batch_mode:
        prompt_save_config(config, args.config or CONFIG_FILENAME)


if __name__ == "__main__":
    main(sys.argv[1:])



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Mapping
from datetime import datetime
from contextlib import contextmanager
from importlib import import_module
from itertools import count
from logging import getLogger

import getpass
import os
import socket
import time
import uuid

from . import exceptions
from . import internalmigrations
from . import utils
from .migrations import topological_sort

logger = getLogger("yoyo.migrations")


class TransactionManager(object):
    """
    Returned by the :meth:`~yoyo.backends.DatabaseBackend.transaction`
    context manager.

    If rollback is called, the transaction is flagged to be rolled back
    when the context manager block closes
    """

    def __init__(self, backend):
        self.backend = backend
        self._rollback = False

    def __enter__(self):
        self._do_begin()
        return self

    def __exit__(self, exc_type, value, traceback):
        if exc_type:
            self._do_rollback()
            return None

        if self._rollback:
            self._do_rollback()
        else:
            self._do_commit()

    def rollback(self):
        """
        Flag that the transaction will be rolled back when the with statement
        exits
        """
        self._rollback = True

    def _do_begin(self):
        """
        Instruct the backend to begin a transaction
        """
        self.backend.begin()

    def _do_commit(self):
        """
        Instruct the backend to commit the transaction
        """
        self.backend.commit()

    def _do_rollback(self):
        """
        Instruct the backend to roll back the transaction
        """
        self.backend.rollback()


class SavepointTransactionManager(TransactionManager):

    id = None
    id_generator = count(1)

    def _do_begin(self):
        assert self.id is None
        self.id = "sp_{}".format(next(self.id_generator))
        self.backend.savepoint(self.id)

    def _do_commit(self):
        """
        This does nothing.

        Trying to the release savepoint here could cause an database error in
        databases where DDL queries cause the transaction to be committed
        and all savepoints released.
        """

    def _do_rollback(self):
        self.backend.savepoint_rollback(self.id)


class DatabaseBackend(object):

    driver_module = ""  # type: str

    log_table = "_yoyo_log"
    lock_table = "yoyo_lock"
    list_tables_sql = "SELECT table_name FROM information_schema.tables"
    version_table = "_yoyo_version"
    migration_table = "_yoyo_migrations"
    is_applied_sql = """
        SELECT COUNT(1) FROM {0.migration_table_quoted}
        WHERE id=:id"""
    mark_migration_sql = (
        "INSERT INTO {0.migration_table_quoted} "
        "(migration_hash, migration_id, applied_at_utc) "
        "VALUES (:migration_hash, :migration_id, :when)"
    )
    unmark_migration_sql = (
        "DELETE FROM {0.migration_table_quoted} WHERE "
        "migration_hash = :migration_hash"
    )
    applied_migrations_sql = (
        "SELECT migration_hash FROM "
        "{0.migration_table_quoted} "
        "ORDER by applied_at_utc"
    )
    create_test_table_sql = (
        "CREATE TABLE {table_name_quoted} " "(id INT PRIMARY KEY)"
    )
    log_migration_sql = (
        "INSERT INTO {0.log_table_quoted} "
        "(id, migration_hash, migration_id, operation, "
        "username, hostname, created_at_utc) "
        "VALUES (:id, :migration_hash, :migration_id, "
        ":operation, :username, :hostname, :created_at_utc)"
    )
    create_lock_table_sql = (
        "CREATE TABLE {0.lock_table_quoted} ("
        "locked INT DEFAULT 1, "
        "ctime TIMESTAMP,"
        "pid INT NOT NULL,"
        "PRIMARY KEY (locked))"
    )

    _driver = None
    _is_locked = False
    _in_transaction = False
    _internal_schema_updated = False

    def __init__(self, dburi, migration_table):
        self.uri = dburi
        self.DatabaseError = self.driver.DatabaseError
        self._connection = self.connect(dburi)
        self.init_connection(self._connection)
        self.migration_table = migration_table
        self.create_lock_table()
        self.has_transactional_ddl = self._check_transactional_ddl()

    def _load_driver_module(self):
        """
        Load the dbapi driver module and register the base exception class
        """
        driver = get_dbapi_module(self.driver_module)
        exceptions.register(driver.DatabaseError)
        return driver

    @property
    def driver(self):
        if self._driver:
            return self._driver
        self._driver = self._load_driver_module()
        return self._driver

    @property
    def connection(self):
        return self._connection

    def init_connection(self, connection):
        """
        Called when creating a connection or after a rollback. May do any
        db specific tasks required to make the connection ready for use.
        """

    def __getattr__(self, attrname):
        if attrname.endswith("_quoted"):
            unquoted = getattr(self, attrname.rsplit("_quoted")[0])
            return self.quote_identifier(unquoted)
        raise AttributeError(attrname)

    def connect(self, dburi):
        raise NotImplementedError()

    def quote_identifier(self, s):
        return '"{}"'.format(s)

    def _check_transactional_ddl(self):
        """
        Return True if the database supports committing/rolling back
        DDL statements within a transaction
        """
        table_name = "yoyo_tmp_{}".format(utils.get_random_string(10))
        table_name_quoted = self.quote_identifier(table_name)
        sql = self.create_test_table_sql.format(
            table_name_quoted=table_name_quoted
        )
        with self.transaction() as t:
            self.execute(sql)
            t.rollback()
        try:
            with self.transaction():
                self.execute("DROP TABLE {}".format(table_name_quoted))
        except self.DatabaseError:
            return True
        return False

    def list_tables(self, **kwargs):
        """
        Return a list of tables present in the backend.
        This is used by the test suite to clean up tables
        generated during testing
        """
        cursor = self.execute(
            self.list_tables_sql,
            dict({"database": self.uri.database}, **kwargs),
        )
        return [row[0] for row in cursor.fetchall()]

    def transaction(self):
        if not self._in_transaction:
            return TransactionManager(self)

        else:
            return SavepointTransactionManager(self)

    def cursor(self):
        return self.connection.cursor()

    def commit(self):
        self.connection.commit()
        self._in_transaction = False

    def rollback(self):
        self.connection.rollback()
        self.init_connection(self.connection)
        self._in_transaction = False

    def begin(self):
        """
        Begin a new transaction
        """
        self._in_transaction = True
        self.execute("BEGIN")

    def savepoint(self, id):
        """
        Create a new savepoint with the given id
        """
        self.execute("SAVEPOINT {}".format(id))

    def savepoint_release(self, id):
        """
        Release (commit) the savepoint with the given id
        """
        self.execute("RELEASE SAVEPOINT {}".format(id))

    def savepoint_rollback(self, id):
        """
        Rollback the savepoint with the given id
        """
        self.execute("ROLLBACK TO SAVEPOINT {}".format(id))

    @contextmanager
    def disable_transactions(self):
        """
        Disable the connection's transaction support, for example by
        setting the isolation mode to 'autocommit'
        """
        self.rollback()
        yield

    @contextmanager
    def lock(self, timeout=10):
        """
        Create a lock to prevent concurrent migrations.

        :param timeout: duration in seconds before raising a LockTimeout error.
        """
        if self._is_locked:
            yield
            return

        pid = os.getpid()
        self._insert_lock_row(pid, timeout)
        try:
            self._is_locked = True
            yield
            self._is_locked = False
        finally:
            self._delete_lock_row(pid)

    def _insert_lock_row(self, pid, timeout, poll_interval=0.5):
        poll_interval = min(poll_interval, timeout)
        started = time.time()
        while True:
            try:
                with self.transaction():
                    self.execute(
                        "INSERT INTO {} (locked, ctime, pid) "
                        "VALUES (1, :when, :pid)".format(
                            self.lock_table_quoted
                        ),
                        {"when": datetime.utcnow(), "pid": pid},
                    )
            except self.DatabaseError:
                if timeout and time.time() > started + timeout:
                    cursor = self.execute(
                        "SELECT pid FROM {}".format(self.lock_table_quoted)
                    )
                    row = cursor.fetchone()
                    if row:
                        raise exceptions.LockTimeout(
                            "Process {} has locked this database "
                            "(run yoyo break-lock to remove this lock)".format(
                                row[0]
                            )
                        )
                    else:
                        raise exceptions.LockTimeout(
                            "Database locked "
                            "(run yoyo break-lock to remove this lock)"
                        )
                time.sleep(poll_interval)
            else:
                return

    def _delete_lock_row(self, pid):
        with self.transaction():
            self.execute(
                "DELETE FROM {} WHERE pid=:pid".format(self.lock_table_quoted),
                {"pid": pid},
            )

    def break_lock(self):
        with self.transaction():
            self.execute("DELETE FROM {}".format(self.lock_table_quoted))

    def execute(self, sql, params=None):
        """
        Create a new cursor, execute a single statement and return the cursor
        object.

        :param sql: A single SQL statement, optionally with named parameters
                    (eg 'SELECT * FROM foo WHERE :bar IS NULL')
        :param params: A dictionary of parameters
        """
        if params and not isinstance(params, Mapping):
            raise TypeError("Expected dict or other mapping object")

        cursor = self.cursor()
        sql, params = utils.change_param_style(
            self.driver.paramstyle, sql, params
        )
        cursor.execute(sql, params)
        return cursor

    def create_lock_table(self):
        """
        Create the lock table if it does not already exist.
        """
        try:
            with self.transaction():
                self.execute(self.create_lock_table_sql.format(self))
        except self.DatabaseError:
            pass

    def ensure_internal_schema_updated(self):
        """
        Check and upgrade yoyo's internal schema.
        """
        if self._internal_schema_updated:
            return
        if internalmigrations.needs_upgrading(self):
            assert not self._in_transaction
            with self.lock():
                internalmigrations.upgrade(self)
                self.connection.commit()
                self._internal_schema_updated = True

    def is_applied(self, migration):
        return migration.hash in self.get_applied_migration_hashes()

    def get_applied_migration_hashes(self):
        """
        Return the list of migration hashes in the order in which they
        were applied
        """
        self.ensure_internal_schema_updated()
        sql = self.applied_migrations_sql.format(self)
        return [row[0] for row in self.execute(sql).fetchall()]

    def to_apply(self, migrations):
        """
        Return the subset of migrations not already applied.
        """
        applied = self.get_applied_migration_hashes()
        ms = (m for m in migrations if m.hash not in applied)
        return migrations.__class__(topological_sort(ms), migrations.post_apply)

    def to_rollback(self, migrations):
        """
        Return the subset of migrations already applied and which may be
        rolled back.

        The order of migrations will be reversed.
        """
        applied = self.get_applied_migration_hashes()
        ms = (m for m in migrations if m.hash in applied)
        return migrations.__class__(
            reversed(list(topological_sort(ms))), migrations.post_apply
        )

    def apply_migrations(self, migrations, force=False):
        if migrations:
            self.apply_migrations_only(migrations, force=force)
            self.run_post_apply(migrations, force=force)

    def apply_migrations_only(self, migrations, force=False):
        """
        Apply the list of migrations, but do not run any post-apply hooks
        present.
        """
        if not migrations:
            return
        for m in migrations:
            try:
                self.apply_one(m, force=force)
            except exceptions.BadMigration:
                continue

    def run_post_apply(self, migrations, force=False):
        """
        Run any post-apply migrations present in ``migrations``
        """
        for m in migrations.post_apply:
            self.apply_one(m, mark=False, force=force)

    def rollback_migrations(self, migrations, force=False):
        self.ensure_internal_schema_updated()
        if not migrations:
            return
        for m in migrations:
            try:
                self.rollback_one(m, force)
            except exceptions.BadMigration:
                continue

    def mark_migrations(self, migrations):
        self.ensure_internal_schema_updated()
        with self.transaction():
            for m in migrations:
                try:
                    self.mark_one(m)
                except exceptions.BadMigration:
                    continue

    def unmark_migrations(self, migrations):
        self.ensure_internal_schema_updated()
        with self.transaction():
            for m in migrations:
                try:
                    self.unmark_one(m)
                except exceptions.BadMigration:
                    continue

    def apply_one(self, migration, force=False, mark=True):
        """
        Apply a single migration
        """
        logger.info("Applying %s", migration.id)
        self.ensure_internal_schema_updated()
        migration.process_steps(self, "apply", force=force)
        self.log_migration(migration, "apply")
        if mark:
            with self.transaction():
                self.mark_one(migration, log=False)

    def rollback_one(self, migration, force=False):
        """
        Rollback a single migration
        """
        logger.info("Rolling back %s", migration.id)
        self.ensure_internal_schema_updated()
        migration.process_steps(self, "rollback", force=force)
        self.log_migration(migration, "rollback")
        with self.transaction():
            self.unmark_one(migration, log=False)

    def unmark_one(self, migration, log=True):
        self.ensure_internal_schema_updated()
        sql = self.unmark_migration_sql.format(self)
        self.execute(sql, {"migration_hash": migration.hash})
        if log:
            self.log_migration(migration, "unmark")

    def mark_one(self, migration, log=True):
        self.ensure_internal_schema_updated()
        logger.info("Marking %s applied", migration.id)
        sql = self.mark_migration_sql.format(self)
        self.execute(
            sql,
            {
                "migration_hash": migration.hash,
                "migration_id": migration.id,
                "when": datetime.utcnow(),
            },
        )
        if log:
            self.log_migration(migration, "mark")

    def log_migration(self, migration, operation, comment=None):
        sql = self.log_migration_sql.format(self)
        self.execute(sql, self.get_log_data(migration, operation, comment))

    def get_log_data(self, migration=None, operation="apply", comment=None):
        """
        Return a dict of data for insertion into the ``_yoyo_log`` table
        """
        assert operation in {"apply", "rollback", "mark", "unmark"}
        return {
            "id": str(uuid.uuid1()),
            "migration_id": migration.id if migration else None,
            "migration_hash": migration.hash if migration else None,
            "username": getpass.getuser(),
            "hostname": socket.getfqdn(),
            "created_at_utc": datetime.utcnow(),
            "operation": operation,
            "comment": comment,
        }


class ODBCBackend(DatabaseBackend):
    driver_module = "pyodbc"

    def connect(self, dburi):
        args = [
            ("UID", dburi.username),
            ("PWD", dburi.password),
            ("ServerName", dburi.hostname),
            ("Port", dburi.port),
            ("Database", dburi.database),
        ]
        args.extend(dburi.args.items())
        s = ";".join("{}={}".format(k, v) for k, v in args if v is not None)
        return self.driver.connect(s)


class OracleBackend(DatabaseBackend):

    driver_module = "cx_Oracle"
    list_tables_sql = "SELECT table_name FROM all_tables WHERE owner=user"

    def begin(self):
        """Oracle is always in a transaction, and has no "BEGIN" statement."""
        self._in_transaction = True

    def connect(self, dburi):
        kwargs = dburi.args
        if dburi.username is not None:
            kwargs["user"] = dburi.username
        if dburi.password is not None:
            kwargs["password"] = dburi.password
        # Oracle combines the hostname, port and database into a single DSN.
        # The DSN can also be a "net service name"
        kwargs["dsn"] = ""
        if dburi.hostname is not None:
            kwargs["dsn"] = dburi.hostname
        if dburi.port is not None:
            kwargs["dsn"] += ":{0}".format(dburi.port)
        if dburi.database is not None:
            if kwargs["dsn"]:
                kwargs["dsn"] += "/{0}".format(dburi.database)
            else:
                kwargs["dsn"] = dburi.database

        return self.driver.connect(**kwargs)


class MySQLBackend(DatabaseBackend):

    driver_module = "pymysql"
    list_tables_sql = (
        "SELECT table_name FROM information_schema.tables "
        "WHERE table_schema = :database"
    )

    def connect(self, dburi):
        kwargs = {"db": dburi.database}
        kwargs.update(dburi.args)
        if dburi.username is not None:
            kwargs["user"] = dburi.username
        if dburi.password is not None:
            kwargs["passwd"] = dburi.password
        if dburi.hostname is not None:
            kwargs["host"] = dburi.hostname
        if dburi.port is not None:
            kwargs["port"] = dburi.port
        if "unix_socket" in dburi.args:
            kwargs["unix_socket"] = dburi.args["unix_socket"]
        if "ssl" in dburi.args:
            kwargs["ssl"] = {}

            if "sslca" in dburi.args:
                kwargs["ssl"]["ca"] = dburi.args["sslca"]

            if "sslcapath" in dburi.args:
                kwargs["ssl"]["capath"] = dburi.args["sslcapath"]

            if "sslcert" in dburi.args:
                kwargs["ssl"]["cert"] = dburi.args["sslcert"]

            if "sslkey" in dburi.args:
                kwargs["ssl"]["key"] = dburi.args["sslkey"]

            if "sslcipher" in dburi.args:
                kwargs["ssl"]["cipher"] = dburi.args["sslcipher"]

        kwargs["db"] = dburi.database
        return self.driver.connect(**kwargs)

    def quote_identifier(self, identifier):
        sql_mode = self.execute("SHOW VARIABLES LIKE 'sql_mode'").fetchone()[1]
        if "ansi_quotes" in sql_mode.lower():
            return super(MySQLBackend, self).quote_identifier(identifier)
        return "`{}`".format(identifier)


class MySQLdbBackend(MySQLBackend):
    driver_module = "MySQLdb"


class SQLiteBackend(DatabaseBackend):

    driver_module = "sqlite3"
    list_tables_sql = "SELECT name FROM sqlite_master WHERE type = 'table'"

    def connect(self, dburi):
        conn = self.driver.connect(
            dburi.database, detect_types=self.driver.PARSE_DECLTYPES
        )
        conn.isolation_level = None
        return conn


class PostgresqlBackend(DatabaseBackend):

    driver_module = "psycopg2"
    schema = None
    list_tables_sql = (
        "SELECT table_name FROM information_schema.tables "
        "WHERE table_schema = :schema"
    )

    def connect(self, dburi):
        kwargs = {"dbname": dburi.database}
        kwargs.update(dburi.args)
        if dburi.username is not None:
            kwargs["user"] = dburi.username
        if dburi.password is not None:
            kwargs["password"] = dburi.password
        if dburi.port is not None:
            kwargs["port"] = dburi.port
        if dburi.hostname is not None:
            kwargs["host"] = dburi.hostname
        self.schema = kwargs.pop("schema", None)
        return self.driver.connect(**kwargs)

    @contextmanager
    def disable_transactions(self):
        with super(PostgresqlBackend, self).disable_transactions():
            saved = self.connection.autocommit
            self.connection.autocommit = True
            yield
            self.connection.autocommit = saved

    def init_connection(self, connection):
        if self.schema:
            cursor = connection.cursor()
            cursor.execute("SET search_path TO {}".format(self.schema))

    def list_tables(self):
        current_schema = self.execute("SELECT current_schema").fetchone()[0]
        return super(PostgresqlBackend, self).list_tables(schema=current_schema)


class RedshiftBackend(PostgresqlBackend):
    def list_tables(self):
        current_schema = self.execute("SELECT current_schema()").fetchone()[0]
        return super(PostgresqlBackend, self).list_tables(schema=current_schema)

    # Redshift does not support ROLLBACK TO SAVEPOINT
    def savepoint(self, id):
        pass

    def savepoint_release(self, id):
        pass

    def savepoint_rollback(self, id):
        self.rollback()

    # Redshift does not enforce primary and unique keys
    def _insert_lock_row(self, pid, timeout, poll_interval=0.5):
        poll_interval = min(poll_interval, timeout)
        started = time.time()
        while True:
            with self.transaction():
                # prevents isolation violation errors
                self.execute("LOCK {}".format(self.lock_table_quoted))
                cursor = self.execute(
                    "SELECT pid FROM {}".format(self.lock_table_quoted)
                )
                row = cursor.fetchone()
                if not row:
                    self.execute(
                        "INSERT INTO {} (locked, ctime, pid) "
                        "VALUES (1, :when, :pid)".format(
                            self.lock_table_quoted
                        ),
                        {"when": datetime.utcnow(), "pid": pid},
                    )
                    return
                elif timeout and time.time() > started + timeout:
                    raise exceptions.LockTimeout(
                        "Process {} has locked this database "
                        "(run yoyo break-lock to remove this lock)".format(
                            row[0]
                        )
                    )
                else:
                    time.sleep(poll_interval)


class SnowflakeBackend(DatabaseBackend):

    driver_module = "snowflake.connector"

    def connect(self, dburi):
        database, schema = dburi.database.split("/")
        return self.driver.connect(
            user=dburi.username,
            password=dburi.password,
            account=dburi.hostname,
            database=database,
            schema=schema,
            warehouse=dburi.args["warehouse"],
        )

    def savepoint(self, id):
        pass

    def savepoint_release(self, id):
        pass

    def savepoint_rollback(self, id):
        pass


def get_dbapi_module(name):
    """
    Import and return the named DB-API driver module
    """
    return import_module(name)



"""
Version 1 schema.

Compatible with yoyo-migrations < 6.0
"""


def upgrade(backend):
    create_migration_table(backend)


def create_migration_table(backend):
    backend.execute(
        "CREATE TABLE {0.migration_table_quoted} ("
        "id VARCHAR(191) NOT NULL PRIMARY KEY,"
        "ctime TIMESTAMP)".format(backend)
    )



from tempfile import NamedTemporaryFile
from threading import Thread
import time

from mock import Mock
from mock import call
from mock import patch
import pytest

from yoyo import backends
from yoyo import read_migrations
from yoyo import exceptions
from yoyo.connections import get_backend
from yoyo.tests import get_test_backends
from yoyo.tests import get_test_dburis
from yoyo.tests import with_migrations


class TestTransactionHandling(object):
    def test_it_commits(self, backend):
        with backend.transaction():
            backend.execute("INSERT INTO yoyo_t values ('A')")

        with backend.transaction():
            rows = list(backend.execute("SELECT * FROM yoyo_t").fetchall())
            assert rows == [("A",)]

    def test_it_rolls_back(self, backend):
        with pytest.raises(backend.DatabaseError):
            with backend.transaction():
                backend.execute("INSERT INTO yoyo_t values ('A')")
                # Invalid SQL to produce an error
                backend.execute("INSERT INTO nonexistant values ('A')")

        with backend.transaction():
            rows = list(backend.execute("SELECT * FROM yoyo_t").fetchall())
            assert rows == []

    def test_it_nests_transactions(self, backend):
        if "redshift" in backend.uri.scheme:
            pytest.skip("Nested transactions not supported for Redshift")

        with backend.transaction():
            backend.execute("INSERT INTO yoyo_t values ('A')")

            with backend.transaction() as trans:
                backend.execute("INSERT INTO yoyo_t values ('B')")
                trans.rollback()

            with backend.transaction() as trans:
                backend.execute("INSERT INTO yoyo_t values ('C')")

        with backend.transaction():
            rows = list(backend.execute("SELECT * FROM yoyo_t").fetchall())
            assert rows == [("A",), ("C",)]

    def test_redshift_nested_transactions(self, backend):
        if "redshift" not in backend.uri.scheme:
            pytest.skip("Redshift only test")

        with backend.transaction():
            backend.execute("INSERT INTO yoyo_t values ('A')")

            with backend.transaction() as trans:
                backend.execute("INSERT INTO yoyo_t values ('B')")
                trans.rollback()

            with backend.transaction() as trans:
                backend.execute("INSERT INTO yoyo_t values ('C')")

        with backend.transaction():
            rows = list(backend.execute("SELECT * FROM yoyo_t").fetchall())
            assert rows == [("C",)]

    def test_backend_detects_transactional_ddl(self, backend):
        expected = {
            backends.PostgresqlBackend: True,
            backends.RedshiftBackend: True,
            backends.SQLiteBackend: True,
            backends.MySQLBackend: False,
        }
        if backend.__class__ in expected:
            assert backend.has_transactional_ddl is expected[backend.__class__]

    def test_non_transactional_ddl_behaviour(self, backend):
        """
        DDL queries in MySQL commit the current transaction,
        but it still seems to respect a subsequent rollback.

        We don't rely on this behaviour, but it's weird and worth having
        a test to document how it works and flag up in future should a new
        backend do things differently
        """
        if backend.has_transactional_ddl:
            return

        with backend.transaction() as trans:
            backend.execute("CREATE TABLE yoyo_a (id INT)")  # implicit commit
            backend.execute("INSERT INTO yoyo_a VALUES (1)")
            backend.execute("CREATE TABLE yoyo_b (id INT)")  # implicit commit
            backend.execute("INSERT INTO yoyo_b VALUES (1)")
            trans.rollback()

        count_a = backend.execute("SELECT COUNT(1) FROM yoyo_a").fetchall()[0][
            0
        ]
        assert count_a == 1

        count_b = backend.execute("SELECT COUNT(1) FROM yoyo_b").fetchall()[0][
            0
        ]
        assert count_b == 0

    @with_migrations(
        a="""
        __transactional__ = False
        step('CREATE DATABASE yoyo_test_tmp',
             'DROP DATABASE yoyo_test_tmp',
             )
    """
    )
    def test_statements_requiring_no_transaction(self, tmpdir):
        """
        PostgreSQL will error if certain statements (eg CREATE DATABASE)
        are run within a transaction block.

        As far as I know this behavior is PostgreSQL specific. We can't run
        this test in sqlite or oracle as they do not support CREATE DATABASE.
        """
        for backend in get_test_backends(exclude={"sqlite", "oracle"}):
            migrations = read_migrations(tmpdir)
            backend.apply_migrations(migrations)
            backend.rollback_migrations(migrations)

    @with_migrations(
        a="""
        __transactional__ = False
        def reopen_db(conn):
            import sqlite3
            for _, db, filename in conn.execute('PRAGMA database_list'):
                if db == 'main':
                     reconn = sqlite3.connect(filename)
                     reconn.execute("CREATE TABLE yoyo_test_b (id int)")
                     break
            else:
                raise AssertionError("sqlite main database not found")

        step('CREATE TABLE yoyo_test_a (id int)')
        step(reopen_db)
        step('CREATE TABLE yoyo_test_c (id int)')
    """
    )
    def test_disabling_transactions_in_sqlite(self, tmpdir):
        """
        Transactions cause sqlite databases to become locked, preventing
        other tools from accessing them:

        https://bitbucket.org/ollyc/yoyo/issues/43/run-step-outside-of-transaction
        """
        with NamedTemporaryFile() as tmp:
            backend = get_backend("sqlite:///" + tmp.name)
            backend.apply_migrations(read_migrations(tmpdir))
            assert "yoyo_test_a" in backend.list_tables()
            assert "yoyo_test_b" in backend.list_tables()
            assert "yoyo_test_c" in backend.list_tables()


class TestConcurrency(object):

    # How long to lock for: long enough to allow a migration to be loaded and
    # started without unduly slowing down the test suite
    lock_duration = 0.5

    def get_lock_sleeper(self, dburi):
        """
        Return a function that acquires the backend lock, then sleeps
        """
        backend = get_backend(dburi)

        def lock_sleep():
            with backend.lock():
                time.sleep(self.lock_duration)

        return lock_sleep

    def skip_if_not_concurrency_safe(self, backend):
        if (
            "sqlite" in backend.uri.scheme
            and backend.uri.database == ":memory:"
        ):
            pytest.skip(
                "Concurrency tests not supported for SQLite "
                "in-memory databases, which cannot be shared "
                "between threads"
            )
        if backend.driver.threadsafety < 1:
            pytest.skip(
                "Concurrency tests not supported for non-threadsafe backends"
            )

    def test_lock(self, dburi):
        """
        Test that :meth:`~yoyo.backends.DatabaseBackend.lock`
        acquires an exclusive lock
        """
        backend = get_backend(dburi)
        self.skip_if_not_concurrency_safe(backend)
        thread = Thread(target=self.get_lock_sleeper(dburi))
        t = time.time()
        thread.start()

        # Give the thread time to acquire the lock, but not enough
        # to complete
        time.sleep(self.lock_duration * 0.6)

        with backend.lock():
            delta = time.time() - t
            assert delta >= self.lock_duration

        thread.join()

    def test_lock_times_out(self, dburi):

        backend = get_backend(dburi)
        self.skip_if_not_concurrency_safe(backend)

        thread = Thread(target=self.get_lock_sleeper(dburi))
        thread.start()
        # Give the thread time to acquire the lock, but not enough
        # to complete
        time.sleep(self.lock_duration * 0.6)
        with pytest.raises(exceptions.LockTimeout):
            with backend.lock(timeout=0.001):
                assert False, "Execution should never reach this point"

        thread.join()


class TestInitConnection(object):
    class MockBackend(backends.DatabaseBackend):
        driver = Mock(DatabaseError=Exception, paramstyle="format")

        def list_tables(self):
            return []

        def connect(self, dburi):
            return Mock()

    def test_it_calls_init_connection(self):

        with patch("yoyo.internalmigrations.upgrade"), patch.object(
            self.MockBackend, "init_connection", Mock()
        ) as mock_init:

            backend = self.MockBackend("", "")
            connection = backend.connection
            assert mock_init.call_args == call(connection)

            mock_init.reset_mock()
            backend.rollback()
            assert mock_init.call_args_list == [call(connection)]

    def test_postgresql_backend_sets_search_path(self):
        class MockPGBackend(backends.PostgresqlBackend):
            driver = Mock(DatabaseError=Exception, paramstyle="format")
            schema = "foo"

            def connect(self, dburi):
                return Mock()

        with patch("yoyo.internalmigrations.upgrade"):
            backend = MockPGBackend("", "")
            backend.rollback()
            assert backend.connection.cursor().execute.call_args == call(
                "SET search_path TO foo"
            )

    def test_postgresql_connects_with_schema(self):
        dburi = next(iter(get_test_dburis(only={"postgresql"})), None)
        if dburi is None:
            pytest.skip("PostgreSQL backend not available")
            return
        backend = get_backend(dburi)
        with backend.transaction():
            backend.execute("CREATE SCHEMA foo")
        try:
            assert get_backend(dburi + "?schema=foo").execute(
                "SHOW search_path"
            ).fetchone() == ("foo",)
        finally:
            with backend.transaction():
                backend.execute("DROP SCHEMA foo CASCADE")

    def test_postgresql_list_table_uses_current_schema(self):
        dburi = next(iter(get_test_dburis(only={"postgresql"})), None)
        if dburi is None:
            pytest.skip("PostgreSQL backend not available")
        backend = get_backend(dburi)
        dbname = backend.uri.database
        with backend.transaction():
            backend.execute(
                "ALTER DATABASE {} SET SEARCH_PATH = custom_schema,public".format(
                    dbname
                )
            )
        try:
            with backend.transaction():
                backend.execute("CREATE SCHEMA custom_schema")
                backend.execute("CREATE TABLE custom_schema.foo (x int)")
            assert "foo" in get_backend(dburi).list_tables()

        finally:
            with backend.transaction():
                backend.execute(
                    "ALTER DATABASE {} RESET SEARCH_PATH".format(dbname)
                )
                backend.execute("DROP SCHEMA custom_schema CASCADE")



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from datetime import timedelta
from mock import Mock, patch
import io
import itertools
import os

import pytest

from yoyo.connections import get_backend
from yoyo import read_migrations
from yoyo import exceptions
from yoyo import ancestors, descendants

from yoyo.tests import with_migrations, migrations_dir, dburi
from yoyo.tests import tempdir
from yoyo.migrations import topological_sort, MigrationList
from yoyo.scripts import newmigration


@with_migrations(
    """
    step("CREATE TABLE yoyo_test (id INT)")
    """,
    """
step("INSERT INTO yoyo_test VALUES (1)")
step("INSERT INTO yoyo_test VALUES ('x', 'y')")
    """,
)
def test_transaction_is_not_committed_on_error(tmpdir):
    backend = get_backend(dburi)
    migrations = read_migrations(tmpdir)
    with pytest.raises(backend.DatabaseError):
        backend.apply_migrations(migrations)
    cursor = backend.cursor()
    cursor.execute("SELECT count(1) FROM yoyo_test")
    assert cursor.fetchone() == (0,)


@with_migrations(
    'step("CREATE TABLE yoyo_test (id INT)")',
    """
step("INSERT INTO yoyo_test VALUES (1)", "DELETE FROM yoyo_test WHERE id=1")
step("UPDATE yoyo_test SET id=2 WHERE id=1", "UPDATE yoyo_test SET id=1 WHERE id=2")
    """,
)
def test_rollbacks_happen_in_reverse(tmpdir):
    backend = get_backend(dburi)
    migrations = read_migrations(tmpdir)
    backend.apply_migrations(migrations)
    cursor = backend.cursor()
    cursor.execute("SELECT * FROM yoyo_test")
    assert cursor.fetchall() == [(2,)]
    backend.rollback_migrations(migrations)
    cursor.execute("SELECT * FROM yoyo_test")
    assert cursor.fetchall() == []


@with_migrations(
    """
    step("CREATE TABLE yoyo_test (id INT)")
    step("INSERT INTO yoyo_test VALUES (1)")
    step("INSERT INTO yoyo_test VALUES ('a', 'b')", ignore_errors='all')
    step("INSERT INTO yoyo_test VALUES (2)")
    """
)
def test_execution_continues_with_ignore_errors(tmpdir):
    backend = get_backend(dburi)
    migrations = read_migrations(tmpdir)
    backend.apply_migrations(migrations)
    cursor = backend.cursor()
    cursor.execute("SELECT * FROM yoyo_test")
    assert cursor.fetchall() == [(1,), (2,)]


@with_migrations(
    """
    from yoyo import step, group
    step("CREATE TABLE yoyo_test (id INT)")
    group(
        step("INSERT INTO yoyo_test VALUES (1)"),
        step("INSERT INTO yoyo_test VALUES ('a', 'b')"),
        ignore_errors='all'
    )
    step("INSERT INTO yoyo_test VALUES (2)")
    """
)
def test_execution_continues_with_ignore_errors_in_transaction(tmpdir):
    backend = get_backend(dburi)
    migrations = read_migrations(tmpdir)
    backend.apply_migrations(migrations)
    cursor = backend.cursor()
    cursor.execute("SELECT * FROM yoyo_test")
    assert cursor.fetchall() == [(2,)]


@with_migrations(
    """
    step("CREATE TABLE yoyo_test (id INT)")
    step("INSERT INTO yoyo_test VALUES (1)",
         "DELETE FROM yoyo_test WHERE id=2")
    step("UPDATE yoyo_test SET id=2 WHERE id=1",
         "SELECT nonexistent FROM imaginary", ignore_errors='rollback')
    """
)
def test_rollbackignores_errors(tmpdir):
    backend = get_backend(dburi)
    migrations = read_migrations(tmpdir)
    backend.apply_migrations(migrations)
    cursor = backend.cursor()
    cursor.execute("SELECT * FROM yoyo_test")
    assert cursor.fetchall() == [(2,)]

    backend.rollback_migrations(migrations)
    cursor.execute("SELECT * FROM yoyo_test")
    assert cursor.fetchall() == []


def test_migration_is_committed(backend):
    with migrations_dir('step("CREATE TABLE yoyo_test (id INT)")') as tmpdir:
        migrations = read_migrations(tmpdir)
        backend.apply_migrations(migrations)

    backend.rollback()
    rows = backend.execute("SELECT * FROM yoyo_test").fetchall()
    assert list(rows) == []


def test_rollback_happens_on_step_failure(backend):
    with migrations_dir(
        """
                        step("",
                             "CREATE TABLE yoyo_is_rolledback (i INT)"),
                        step("CREATE TABLE yoyo_test (s VARCHAR(100))",
                             "DROP TABLE yoyo_test")
                        step("invalid sql!")"""
    ) as tmpdir:
        migrations = read_migrations(tmpdir)
        with pytest.raises(backend.DatabaseError):
            backend.apply_migrations(migrations)

    # The yoyo_test table should have either been deleted (transactional ddl)
    # or dropped (non-transactional-ddl)
    with pytest.raises(backend.DatabaseError):
        backend.execute("SELECT * FROM yoyo_test")

    # Transactional DDL: rollback steps not executed
    if backend.has_transactional_ddl:
        with pytest.raises(backend.DatabaseError):
            backend.execute("SELECT * FROM yoyo_is_rolledback")

    # Non-transactional DDL: ensure the rollback steps were executed
    else:
        cursor = backend.execute("SELECT * FROM yoyo_is_rolledback")
        assert list(cursor.fetchall()) == []


@with_migrations(
    """
    step("CREATE TABLE yoyo_test (id INT)")
    step("DROP TABLE yoyo_test")
    """
)
def test_specify_migration_table(tmpdir):
    backend = get_backend(dburi, migration_table="another_migration_table")
    migrations = read_migrations(tmpdir)
    backend.apply_migrations(migrations)
    cursor = backend.cursor()
    cursor.execute("SELECT migration_id FROM another_migration_table")
    assert cursor.fetchall() == [("0",)]


@with_migrations(
    """
    def foo(conn):
        conn.cursor().execute("CREATE TABLE foo_test (id INT)")
        conn.cursor().execute("INSERT INTO foo_test VALUES (1)")
    def bar(conn):
        foo(conn)
    step(bar)
    """
)
def test_migration_functions_have_namespace_access(tmpdir):
    """
    Test that functions called via step have access to the script namespace
    """
    backend = get_backend(dburi)
    migrations = read_migrations(tmpdir)
    backend.apply_migrations(migrations)
    cursor = backend.cursor()
    cursor.execute("SELECT id FROM foo_test")
    assert cursor.fetchall() == [(1,)]


@with_migrations(
    """
    from yoyo import group, step
    step("CREATE TABLE yoyo_test (id INT)")
    group(step("INSERT INTO yoyo_test VALUES (1)")),
    """
)
def test_migrations_can_import_step_and_group(tmpdir):
    backend = get_backend(dburi)
    migrations = read_migrations(tmpdir)
    backend.apply_migrations(migrations)
    cursor = backend.cursor()
    cursor.execute("SELECT id FROM yoyo_test")
    assert cursor.fetchall() == [(1,)]


@with_migrations(
    """
    step("CREATE TABLE yoyo_test (id INT, c VARCHAR(1))")
    step("INSERT INTO yoyo_test VALUES (1, 'a')")
    step("INSERT INTO yoyo_test VALUES (2, 'b')")
    step("SELECT * FROM yoyo_test")
    """
)
def test_migrations_display_selected_data(tmpdir):
    backend = get_backend(dburi)
    migrations = read_migrations(tmpdir)
    with patch("yoyo.migrations.sys.stdout") as stdout:
        backend.apply_migrations(migrations)
        written = "".join(a[0] for a, kw in stdout.write.call_args_list)
        assert written == (
            " id | c \n" "----+---\n" " 1  | a \n" " 2  | b \n" "(2 rows)\n"
        )


def test_grouped_migrations_can_be_rolled_back(backend):
    with with_migrations(
        a="from yoyo import step\n"
        'steps = [step("CREATE TABLE p (n INT)",'
        '              "DROP TABLE p")]'
    ) as t1:
        backend.apply_migrations(read_migrations(t1))

        with with_migrations(
            b="from yoyo import group\n"
            "from yoyo import step\n"
            "steps = [\n"
            "    group(\n"
            '        step("INSERT INTO p VALUES (3)", \n'
            '             "DELETE FROM p WHERE n=3"),\n'
            '        step("UPDATE p SET n = n * 2", \n'
            '             "UPDATE p SET n = n / 2"),\n'
            '        step("UPDATE p SET n = n + 1", \n'
            '             "UPDATE p SET n = n - 1"),\n'
            "   )\n"
            "]\n"
        ) as t2:
            migrations = read_migrations(t2)
            backend.apply_migrations(migrations)
            backend.rollback_migrations(migrations)
            cursor = backend.execute("SELECT count(1) FROM p")
            assert cursor.fetchone() == (0,)
        backend.rollback_migrations(read_migrations(t1))


class TestTopologicalSort(object):
    def check(self, nodes, edges, expected_order):
        migrations = self.get_mock_migrations(nodes, edges)
        output_order = "".join(m.id for m in topological_sort(migrations))
        assert output_order == expected_order

    def get_mock_migrations(self, nodes="ABCD", edges=[]):
        class MockMigration(Mock):
            def __repr__(self):
                return "<MockMigration {}>".format(self.id)

        migrations = {n: MockMigration(id=n, depends=set()) for n in nodes}
        for g in edges:
            for edge in zip(g, g[1:]):
                migrations[edge[1]].depends.add(migrations[edge[0]])

        return [migrations[n] for n in nodes]

    def test_it_keeps_stable_order(self):

        for s in itertools.permutations("ABCD"):
            self.check(s, {}, "".join(s))

    def test_it_sorts_topologically(self):

        # Single group at start
        self.check("ABCD", {"AB"}, "ABCD")
        self.check("BACD", {"AB"}, "ABCD")

        # Single group in middle start
        self.check("CABD", {"AB"}, "CABD")
        self.check("CBAD", {"AB"}, "CABD")

        # Extended group
        self.check("ABCD", {"AB", "AD"}, "ABDC")
        self.check("DBCA", {"AB", "AD"}, "ADBC")

        # Non-connected groups
        self.check("ABCDEF", {"CB", "ED"}, "ACBEDF")
        self.check("ADEBCF", {"CB", "ED"}, "AEDCBF")
        self.check("ADEFBC", {"CB", "ED"}, "AEDFCB")
        self.check("DBAFEC", {"CB", "ED"}, "EDCBAF")

    def test_it_discards_missing_dependencies(self):
        A, B, C, D = self.get_mock_migrations()
        C.depends.add(Mock())
        assert list(topological_sort([A, B, C, D])) == [A, B, C, D]

    def test_it_catches_cycles(self):
        A, B, C, D = self.get_mock_migrations()
        C.depends.add(C)
        with pytest.raises(exceptions.BadMigration):
            self.check("ABCD", {"AA"}, "")
        with pytest.raises(exceptions.BadMigration):
            self.check("ABCD", {"AB", "BA"}, "")
        with pytest.raises(exceptions.BadMigration):
            self.check("ABCD", {"AB", "BC", "CB"}, "")
        with pytest.raises(exceptions.BadMigration):
            self.check("ABCD", {"AB", "BC", "CA"}, "")

    def test_it_handles_multiple_edges_to_the_same_node(self):
        self.check("ABCD", {"AB", "AC", "AD"}, "ABCD")
        self.check("DCBA", {"AB", "AC", "AD"}, "ADCB")

    def test_it_handles_multiple_edges_to_the_same_node2(self):
        #      A --> B
        #      |     ^
        #      v     |
        #      C --- +
        for input_order in itertools.permutations("ABC"):
            self.check(input_order, {"AB", "AC", "CB"}, "ACB")


class TestMigrationList(object):
    def test_can_create_empty(self):
        m = MigrationList()
        assert list(m) == []

    def test_cannot_create_with_duplicate_ids(self):
        with pytest.raises(exceptions.MigrationConflict):
            MigrationList([Mock(id=1), Mock(id=1)])

    def test_can_append_new_id(self):
        m = MigrationList([Mock(id=n) for n in range(10)])
        m.append(Mock(id=10))

    def test_cannot_append_duplicate_id(self):
        m = MigrationList([Mock(id=n) for n in range(10)])
        with pytest.raises(exceptions.MigrationConflict):
            m.append(Mock(id=1))

    def test_deletion_allows_reinsertion(self):
        m = MigrationList([Mock(id=n) for n in range(10)])
        del m[0]
        m.append(Mock(id=0))

    def test_can_overwrite_slice_with_same_ids(self):
        m = MigrationList([Mock(id=n) for n in range(10)])
        m[1:3] = [Mock(id=2), Mock(id=1)]

    def test_cannot_overwrite_slice_with_conflicting_ids(self):
        m = MigrationList([Mock(id=n) for n in range(10)])
        with pytest.raises(exceptions.MigrationConflict):
            m[1:3] = [Mock(id=4)]


class TestAncestorsDescendants(object):
    def setup(self):
        self.m1 = Mock(id="m1", depends=["m2", "m3"])
        self.m2 = Mock(id="m2", depends=["m3"])
        self.m3 = Mock(id="m3", depends=["m5"])
        self.m4 = Mock(id="m4", depends=["m5"])
        self.m5 = Mock(id="m5", depends=[])
        self.m1.depends = {self.m2, self.m3}
        self.m2.depends = {self.m3}
        self.m3.depends = {self.m5}
        self.m4.depends = {self.m5}
        self.migrations = {self.m1, self.m2, self.m3, self.m4, self.m5}

    def test_ancestors(self):

        assert ancestors(self.m1, self.migrations) == {
            self.m2,
            self.m3,
            self.m5,
        }
        assert ancestors(self.m2, self.migrations) == {self.m3, self.m5}
        assert ancestors(self.m3, self.migrations) == {self.m5}
        assert ancestors(self.m4, self.migrations) == {self.m5}
        assert ancestors(self.m5, self.migrations) == set()

    def test_descendants(self):

        assert descendants(self.m1, self.migrations) == set()
        assert descendants(self.m2, self.migrations) == {self.m1}
        assert descendants(self.m3, self.migrations) == {self.m2, self.m1}
        assert descendants(self.m4, self.migrations) == set()
        assert descendants(self.m5, self.migrations) == {
            self.m4,
            self.m3,
            self.m2,
            self.m1,
        }


class TestReadMigrations(object):
    @with_migrations(**{newmigration.tempfile_prefix + "test": ""})
    def test_it_ignores_yoyo_new_tmp_files(self, tmpdir):
        """
        The yoyo new command creates temporary files in the migrations directory.
        These shouldn't be picked up by yoyo apply etc
        """
        assert len(read_migrations(tmpdir)) == 0

    @with_migrations(**{"post-apply": """step('SELECT 1')"""})
    def test_it_loads_post_apply_scripts(self, tmpdir):
        migrations = read_migrations(tmpdir)
        assert len(migrations) == 0
        assert len(migrations.post_apply) == 1

    @with_migrations(**{"a": """step('SELECT 1')"""})
    def test_it_does_not_add_duplicate_steps(self, tmpdir):
        m = read_migrations(tmpdir)[0]
        m.load()
        assert len(m.steps) == 1

        m = read_migrations(tmpdir)[0]
        m.load()
        assert len(m.steps) == 1

    @with_migrations(**{"a": """from yoyo import step; step('SELECT 1')"""})
    def test_it_does_not_add_duplicate_steps_with_imported_symbols(
        self, tmpdir
    ):
        m = read_migrations(tmpdir)[0]
        m.load()
        assert len(m.steps) == 1

        m = read_migrations(tmpdir)[0]
        m.load()
        assert len(m.steps) == 1

    def test_it_reads_from_package_data(self):
        migrations = read_migrations("package:yoyo:tests/migrations")
        assert len(migrations) == 1
        assert migrations[0].id == "test-pkg-migration"

    def test_it_globs_directory_names(self):
        def touch(f):
            io.open(f, "w").close()

        with tempdir() as t:
            os.mkdir(os.path.join(t, "aa"))
            os.mkdir(os.path.join(t, "ab"))
            os.mkdir(os.path.join(t, "b"))
            touch(os.path.join(t, "aa", "1.py"))
            touch(os.path.join(t, "ab", "2.py"))
            touch(os.path.join(t, "b", "3.py"))

            migrations = read_migrations("{}/a*".format(t))
            migration_ids = [m.id for m in migrations]
            assert "1" in migration_ids
            assert "2" in migration_ids
            assert "3" not in migration_ids

    def test_it_loads_sql_migrations(self):
        mdir = migrations_dir(
            **{
                "1.sql": "CREATE TABLE foo (id int)",
                "1.rollback.sql": "DROP TABLE foo",
            }
        )
        with mdir as tmp:
            migrations = read_migrations(tmp)
            assert [m.id for m in migrations] == ["1"]
            m = migrations[0]
            m.load()
            assert m.steps[0].step._apply == "CREATE TABLE foo (id int)"
            assert m.steps[0].step._rollback == "DROP TABLE foo"

    def test_it_sets_transactional_for_sql_migrations(self):
        def check(sql, expected):
            with migrations_dir(**{"1.sql": sql}) as tmp:
                migration = read_migrations(tmp)[0]
                migration.load()
                assert migration.use_transactions is expected
                assert migration.steps[0].step._apply == "SELECT 1"

        check("SELECT 1", True)
        check("-- transactional: true\nSELECT 1", True)
        check("-- transactional: false\nSELECT 1", False)
        check("-- transactional: FALSE\nSELECT 1", False)
        check("-- transactional: FALSE\r\nSELECT 1", False)
        check("-- I like bananas!\n-- transactional: FALSE\nSELECT 1", False)

    def test_it_sets_docstring_for_sql_migrations(self):
        def check(sql, expected):
            with migrations_dir(**{"1.sql": sql}) as tmp:
                migration = read_migrations(tmp)[0]
                migration.load()
                assert migration.module.__doc__ == expected
                assert migration.steps[0].step._apply == "SELECT 1"

        check("SELECT 1", "")
        check("-- foo\n-- transactional: true\n--bar\nSELECT 1", "foo\nbar")

    def test_it_sets_depends_for_sql_migrations(self):
        def check(sql, expected):
            with migrations_dir(
                **{"1.sql": "", "2.sql": "", "3.sql": sql}
            ) as tmp:

                migration = read_migrations(tmp)[-1]
                migration.load()
                assert {m.id for m in migration._depends} == expected
                assert migration.steps[0].step._apply == "SELECT 1"

        check("SELECT 1", set())
        check("-- depends: 1\nSELECT 1", {"1"})
        check("-- depends: 1 2\nSELECT 1", {"1", "2"})
        check("-- depends: 2\n-- depends : 1\nSELECT 1", {"1", "2"})
        with pytest.raises(exceptions.BadMigration):
            check("-- depends: true\nSELECT 1", set())

    def test_it_does_not_mix_up_migrations_from_different_sources(self):
        backend = get_backend(dburi)
        with migrations_dir(**{"1.sql": "", "3.sql": ""}) as t1, migrations_dir(
            **{"2.sql": "", "4.sql": ""}
        ) as t2:
            migrations = read_migrations(t1, t2)
            assert [m.id for m in backend.to_apply(migrations)] == [
                "1",
                "3",
                "2",
                "4",
            ]


class TestPostApplyHooks(object):
    def test_post_apply_hooks_are_run_every_time(self):

        backend = get_backend(dburi)
        migrations = migrations_dir(
            **{
                "a": "step('create table postapply (i int)')",
                "post-apply": "step('insert into postapply values (1)')",
            }
        )

        with migrations as tmp:

            def count_postapply_calls():
                cursor = backend.cursor()
                cursor.execute("SELECT count(1) FROM postapply")
                return cursor.fetchone()[0]

            def _apply_migrations():
                backend.apply_migrations(backend.to_apply(read_migrations(tmp)))

            # Should apply migration 'a' and call the post-apply hook
            _apply_migrations()
            assert count_postapply_calls() == 1

            # No outstanding migrations: post-apply hook should not be called
            _apply_migrations()
            assert count_postapply_calls() == 1

            # New migration added: post-apply should be called a second time
            migrations.add_migration("b", "")
            _apply_migrations()
            assert count_postapply_calls() == 2

    @with_migrations(
        **{
            "a": "step('create table postapply (i int)')",
            "post-apply": "step('insert into postapply values (1)')",
            "post-apply2": "step('insert into postapply values (2)')",
        }
    )
    def test_it_runs_multiple_post_apply_hooks(self, tmpdir):
        backend = get_backend(dburi)
        backend.apply_migrations(backend.to_apply(read_migrations(tmpdir)))
        cursor = backend.cursor()
        cursor.execute("SELECT * FROM postapply")
        assert cursor.fetchall() == [(1,), (2,)]

    @with_migrations(
        **{
            "a": "step('create table postapply (i int)')",
            "post-apply": "step('insert into postapply values (1)')",
        }
    )
    def test_apply_migrations_only_does_not_run_hooks(self, tmpdir):
        backend = get_backend(dburi)
        backend.apply_migrations_only(backend.to_apply(read_migrations(tmpdir)))
        cursor = backend.cursor()
        cursor.execute("SELECT * FROM postapply")
        assert cursor.fetchall() == []


class TestLogging(object):
    def get_last_log_entry(self, backend):
        cursor = backend.execute(
            "SELECT migration_id, operation, "
            "created_at_utc, username, hostname "
            "from _yoyo_log "
            "ORDER BY id DESC LIMIT 1"
        )
        return {
            d[0]: value
            for d, value in zip(cursor.description, cursor.fetchone())
        }

    def get_log_count(self, backend):
        return backend.execute("SELECT count(1) FROM _yoyo_log").fetchone()[0]

    def test_it_logs_apply_and_rollback(self, backend):
        with with_migrations(
            a='step("CREATE TABLE yoyo_test (id INT)")'
        ) as tmpdir:
            migrations = read_migrations(tmpdir)
            backend.apply_migrations(migrations)
            assert self.get_log_count(backend) == 1
            logged = self.get_last_log_entry(backend)
            assert logged["migration_id"] == "a"
            assert logged["operation"] == "apply"
            assert logged["created_at_utc"] >= datetime.utcnow() - timedelta(
                seconds=3
            )
            apply_time = logged["created_at_utc"]

            backend.rollback_migrations(migrations)
            assert self.get_log_count(backend) == 2
            logged = self.get_last_log_entry(backend)
            assert logged["migration_id"] == "a"
            assert logged["operation"] == "rollback"
            assert logged["created_at_utc"] >= apply_time

    def test_it_logs_mark_and_unmark(self, backend):
        with with_migrations(
            a='step("CREATE TABLE yoyo_test (id INT)")'
        ) as tmpdir:
            migrations = read_migrations(tmpdir)
            backend.mark_migrations(migrations)
            assert self.get_log_count(backend) == 1
            logged = self.get_last_log_entry(backend)
            assert logged["migration_id"] == "a"
            assert logged["operation"] == "mark"
            assert logged["created_at_utc"] >= datetime.utcnow() - timedelta(
                seconds=3
            )
            marked_time = logged["created_at_utc"]

            backend.unmark_migrations(migrations)
            assert self.get_log_count(backend) == 2
            logged = self.get_last_log_entry(backend)
            assert logged["migration_id"] == "a"
            assert logged["operation"] == "unmark"
            assert logged["created_at_utc"] >= marked_time



"""
Migrate yoyo's internal table structure
"""

from datetime import datetime

from . import v1
from . import v2


#: Mapping of {schema version number: module}
schema_versions = {0: None, 1: v1, 2: v2}


#: First schema version that supports the yoyo_versions table
USE_VERSION_TABLE_FROM = 2


def needs_upgrading(backend):
    return get_current_version(backend) < max(schema_versions)


def upgrade(backend, version=None):
    """
    Check the currently installed yoyo migrations version and update the
    internal schema
    """
    if version is None:
        desired_version = max(schema_versions)
    else:
        desired_version = version
    current_version = get_current_version(backend)
    with backend.transaction():
        while current_version < desired_version:
            next_version = current_version + 1
            schema_versions[next_version].upgrade(backend)  # type: ignore
            current_version = next_version
            mark_schema_version(backend, current_version)


def get_current_version(backend):
    """
    Return the currently installed yoyo migrations schema version
    """
    tables = set(backend.list_tables())
    version_table = backend.version_table
    if backend.migration_table not in tables:
        return 0
    if version_table not in tables:
        return 1
    with backend.transaction():
        cursor = backend.execute(
            "SELECT max(version) FROM {} ".format(
                backend.quote_identifier(version_table)
            )
        )
        version = cursor.fetchone()[0]
        assert version in schema_versions
        return version


def mark_schema_version(backend, version):
    """
    Mark the given version as having been applied
    """
    assert version in schema_versions
    if version < USE_VERSION_TABLE_FROM:
        return
    backend.execute(
        "INSERT INTO {0.version_table_quoted} VALUES (:version, :when)".format(
            backend
        ),
        {"version": version, "when": datetime.utcnow()},
    )



from configparser import ConfigParser
from typing import Dict
from urllib.parse import quote
import os
import pathlib
import tempfile

import pytest

import yoyo.config


def to_dict(parser: ConfigParser) -> Dict[str, Dict[str, str]]:
    return {section: dict(parser[section]) for section in parser}


def _test_files(files, expected):
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir_path = pathlib.Path(tmpdir)
        for path, data in files.items():
            with (tmpdir_path / path).open("w") as f:
                f.write(data)
        saved = os.getcwd()
        os.chdir(tmpdir)
        try:
            config = yoyo.config.read_config(
                str(tmpdir_path / next(iter(sorted(files))))
            )
        finally:
            os.chdir(saved)
        assert to_dict(config) == expected


class TestInheritance:
    def test_read_config_returns_empty_on_None(self):
        config = yoyo.config.read_config(None)
        assert to_dict(config) == {"DEFAULT": {}}

    def test_inherit_one(self):
        _test_files(
            {
                "a.ini": "[DEFAULT]\n%inherit = b.ini\nx=a\ny=a\n",
                "b.ini": "[DEFAULT]\ny = b\nz = b\n",
            },
            {"DEFAULT": {"x": "a", "y": "a", "z": "b"}},
        )

    def test_inherit_many(self):
        _test_files(
            {
                "a.ini": (
                    "[DEFAULT]\n"
                    "%inherit = b.ini c.ini\n"
                    "v = a\n"
                    "w = a\n"
                    "x = a\n"
                ),
                "b.ini": "[DEFAULT]\nw = b\nx = b\ny = b",
                "c.ini": "[DEFAULT]\nx = c\ny = c\nz = c",
            },
            {"DEFAULT": {"v": "a", "w": "a", "x": "a", "y": "b", "z": "c"}},
        )

    def test_include_one(self):
        _test_files(
            {
                "a.ini": "[DEFAULT]\n%include = b.ini\nx=a\ny=a\n",
                "b.ini": "[DEFAULT]\ny = b\nz = b\n",
            },
            {"DEFAULT": {"x": "a", "y": "b", "z": "b"}},
        )

    def test_include_many(self):
        _test_files(
            {
                "a.ini": (
                    "[DEFAULT]\n"
                    "%include = b.ini c.ini\n"
                    "v = a\n"
                    "w = a\n"
                    "x = a\n"
                ),
                "b.ini": "[DEFAULT]\nw = b\nx = b\ny = b",
                "c.ini": "[DEFAULT]\nx = c\ny = c\nz = c",
            },
            {"DEFAULT": {"v": "a", "w": "b", "x": "c", "y": "c", "z": "c"}},
        )

    def test_nested_inherit(self):
        _test_files(
            {
                "a.ini": (
                    "[DEFAULT]\n"
                    "%inherit = b.ini\n"
                    "v = a\n"
                    "w = a\n"
                    "x = a\n"
                ),
                "b.ini": (
                    "[DEFAULT]\n"
                    "%inherit = c.ini\n"
                    "w = b\n"
                    "x = b\n"
                    "y = b\n"
                ),
                "c.ini": "[DEFAULT]\nx = c\ny = c\nz = c",
            },
            {"DEFAULT": {"v": "a", "w": "a", "x": "a", "y": "b", "z": "c"}},
        )

    def test_nested_include(self):
        _test_files(
            {
                "a.ini": (
                    "[DEFAULT]\n"
                    "%include = b.ini\n"
                    "v = a\n"
                    "w = a\n"
                    "x = a\n"
                ),
                "b.ini": (
                    "[DEFAULT]\n"
                    "%include = c.ini\n"
                    "w = b\n"
                    "x = b\n"
                    "y = b\n"
                ),
                "c.ini": "[DEFAULT]\nx = c\ny = c\nz = c",
            },
            {"DEFAULT": {"v": "a", "w": "b", "x": "c", "y": "c", "z": "c"}},
        )

    def test_it_raises_on_not_found(self):
        with pytest.raises(FileNotFoundError):
            _test_files(
                {"a.ini": "[DEFAULT]\n%inherit = b.ini\n"}, {"DEFAULT": {}},
            )

    def test_it_ignores_not_found(self):
        _test_files(
            {"a.ini": "[DEFAULT]\n%inherit = ?b.ini\n"}, {"DEFAULT": {}},
        )

    def test_it_traps_circular_references(self):
        with pytest.raises(yoyo.config.CircularReferenceError):
            _test_files(
                {
                    "a.ini": "[DEFAULT]\n%inherit = b.ini\n",
                    "b.ini": "[DEFAULT]\n%inherit = a.ini\n",
                },
                {"DEFAULT": {}},
            )

    def test_it_traps_deep_circular_references(self):
        with pytest.raises(yoyo.config.CircularReferenceError):
            _test_files(
                {
                    "a.ini": "[DEFAULT]\n%include = b.ini c.ini\n",
                    "b.ini": "[DEFAULT]\n%include = c.ini\n",
                    "c.ini": "[DEFAULT]\n%include = a.ini\n",
                },
                {"DEFAULT": {}},
            )

    def test_it_allows_acyclic_references(self):
        _test_files(
            {
                "a.ini": "[DEFAULT]\n%include = b.ini c.ini\nw = a\nx = a\n",
                "b.ini": "[DEFAULT]\n%include = c.ini\nx = b\ny = b",
                "c.ini": "[DEFAULT]\ny = c\nz = c\n",
            },
            {"DEFAULT": {"w": "a", "x": "b", "y": "c", "z": "c"}},
        )


class TestInterpolation:
    def test_it_interpolates_environment_variables(self):
        os.environ["yoyo_test_env_var"] = "foo"
        try:
            _test_files(
                {"a.ini": "[DEFAULT]\nx=%(yoyo_test_env_var)s"},
                {"DEFAULT": {"x": "foo"}},
            )
        finally:
            del os.environ["yoyo_test_env_var"]

    def test_it_folds_environment_variable_case(self):
        os.environ["YOYO_TEST_ENV_VAR"] = "foo"
        try:
            _test_files(
                {"a.ini": "[DEFAULT]\nx=%(yoyo_test_env_var)s"},
                {"DEFAULT": {"x": "foo"}},
            )
        finally:
            del os.environ["YOYO_TEST_ENV_VAR"]

    def test_it_allows_url_encoded_values(self):
        os.environ["yoyo_test_env_var"] = quote("p@ssword$")
        try:
            _test_files(
                {"a.ini": "[DEFAULT]\nx=%(yoyo_test_env_var)s"},
                {"DEFAULT": {"x": quote("p@ssword$")}},
            )
        finally:
            del os.environ["yoyo_test_env_var"]



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from mock import patch, call, MagicMock
import pytest

from yoyo.connections import parse_uri, BadConnectionURI


class MockDatabaseError(Exception):
    pass


class TestParseURI:
    @pytest.mark.skipif(
        sys.version_info < (2, 7, 4),
        reason="Requires python>=2.7.4 " "(http://bugs.python.org/issue7904)",
    )
    def test_it_parses_all_fields(self):
        parsed = parse_uri("protocol://scott:tiger@server:666/db?k=1")
        assert tuple(parsed) == (
            "protocol",
            "scott",
            "tiger",
            "server",
            666,
            "db",
            {"k": "1"},
        )

    def test_it_parses_escaped_username(self):
        parsed = parse_uri("protocol://scott%40example.org:tiger@localhost/db")
        assert parsed.username == "scott@example.org"

    def test_it_requires_scheme(self):
        with pytest.raises(BadConnectionURI):
            parse_uri("//scott:tiger@localhost/db")

    def test_it_roundtrips(self):
        cases = [
            "proto://scott%40example.org:tiger@localhost/db",
            "proto://localhost/db?a=1+2",
            "proto://localhost/db?a=a%3D1",
        ]
        for case in cases:
            parsed = parse_uri(case)
            assert parsed.uri == case

    def test_it_returns_relative_paths_for_sqlite(self):
        assert parse_uri("sqlite:///foo/bar.db").database == "foo/bar.db"

    def test_it_returns_absolute_paths_for_sqlite(self):
        assert parse_uri("sqlite:////foo/bar.db").database == "/foo/bar.db"

    def test_passwords_with_slashes_dont_break_netloc(self):
        parsed = parse_uri("postgresql://user:a%2Fb@localhost:5432/db")
        assert parsed.netloc == 'user:a%2Fb@localhost:5432'
        assert parsed.port == 5432
        assert parsed.password == 'a/b'


@patch(
    "yoyo.backends.get_dbapi_module",
    return_value=MagicMock(
        DatabaseError=MockDatabaseError, paramstyle="qmark"
    ),
)
def test_connections(get_dbapi_module):

    from yoyo import backends

    u = parse_uri("odbc://scott:tiger@db.example.org:42/northwind?foo=bar")
    cases = [
        (
            backends.ODBCBackend,
            "pyodbc",
            call(
                "UID=scott;PWD=tiger;ServerName=db.example.org;"
                "Port=42;Database=northwind;foo=bar"
            ),
        ),
        (
            backends.MySQLBackend,
            "pymysql",
            call(
                user="scott",
                passwd="tiger",
                host="db.example.org",
                port=42,
                db="northwind",
                foo="bar",
            ),
        ),
        (
            backends.SQLiteBackend,
            "sqlite3",
            call(
                "northwind",
                detect_types=get_dbapi_module("sqlite3").PARSE_DECLTYPES,
            ),
        ),
        (
            backends.PostgresqlBackend,
            "psycopg2",
            call(
                user="scott",
                password="tiger",
                port=42,
                host="db.example.org",
                dbname="northwind",
                foo="bar",
            ),
        ),
    ]
    for cls, driver_module, connect_args in cases:
        cls(u, "_yoyo_migration")
        assert get_dbapi_module.call_args == call(driver_module)
        assert get_dbapi_module().connect.call_args == connect_args



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import date
from textwrap import dedent
from tempfile import NamedTemporaryFile
import configparser
import glob
import logging
import io
import re
import shlex
import subprocess
import sys
import traceback

from yoyo import default_migration_table
from yoyo.config import CONFIG_NEW_MIGRATION_COMMAND_KEY
from yoyo.migrations import read_migrations, heads, Migration
from yoyo import utils
from .main import InvalidArgument

from os import path, stat, unlink, rename

logger = logging.getLogger("yoyo.migrations")

tempfile_prefix = "_tmp_yoyonew"

migration_template = dedent(
    '''\
    """
    {message}
    """

    from yoyo import step

    __depends__ = {{{depends}}}

    steps = [
        step("")
    ]
    '''
)
migration_sql_template = dedent(
    """\
    -- {message}
    -- depends: {depends}

    """
)


def install_argparsers(global_parser, subparsers):
    parser_new = subparsers.add_parser(
        "new", parents=[global_parser], help="Create a new migration"
    )
    parser_new.set_defaults(func=new_migration)
    parser_new.add_argument("--message", "-m", help="Message", default="")
    parser_new.add_argument(
        "--sql", help="Create file in SQL format", action="store_true"
    )
    parser_new.add_argument(
        "--migration-table",
        dest="migration_table",
        action="store",
        default=default_migration_table,
        help="Name of table to use for storing " "migration metadata",
    )
    parser_new.add_argument(
        "sources", nargs="*", help="Source directory of migration scripts"
    )
    parser_new.add_argument(
        "-d",
        "--database",
        default=None,
        help="Database, eg 'sqlite:///path/to/sqlite.db' "
        "or 'postgresql://user@host/db'",
    )


def new_migration(args, config):

    try:
        directory = args.sources[0]
    except IndexError:
        raise InvalidArgument("Please specify a migrations directory")

    message = args.message
    migrations = read_migrations(directory)
    depends = sorted(heads(migrations), key=lambda m: m.id)
    if args.sql:
        template = migration_sql_template
        depends_str = "  ".join(m.id for m in depends)
    else:
        template = migration_template
        depends_str = ", ".join("{!r}".format(m.id) for m in depends)
    migration_source = template.format(message=message, depends=depends_str)

    extension = ".sql" if args.sql else ".py"
    if args.batch_mode:
        p = make_filename(config, directory, message, extension)
        with io.open(p, "w", encoding="UTF-8") as f:
            f.write(migration_source)
    else:
        p = create_with_editor(config, directory, migration_source, extension)
        if p is None:
            return

    try:
        command = config.get("DEFAULT", CONFIG_NEW_MIGRATION_COMMAND_KEY)
        command = [part.format(p) for part in shlex.split(command)]
        logger.info("Running command: %s", " ".join(command))
        subprocess.call(command)
    except configparser.NoOptionError:
        pass

    print("Created file", p)


def slugify(message):
    s = utils.unidecode(message)
    s = re.sub(re.compile(r"[^-a-z0-9]+"), "-", s.lower())
    s = re.compile(r"-{2,}").sub("-", s).strip("-")
    return s


def make_filename(config, directory, message, extension):
    lines = (line.strip() for line in message.split("\n"))
    lines = (line for line in lines if line)
    message = next(lines, None)

    if message:
        slug = "-" + slugify(message)
    else:
        slug = ""

    datestr = date.today().strftime("%Y%m%d")
    number = "01"
    rand = utils.get_random_string(5)

    try:
        prefix = config.get("DEFAULT", "prefix")
    except configparser.NoOptionError:
        prefix = ""

    for p in glob.glob(path.join(directory, "{}{}_*".format(prefix, datestr))):
        n = path.basename(p)[len(prefix) + len(datestr) + 1 :].split("_")[0]

        try:
            if number <= n:
                number = str(int(n) + 1).zfill(2)
        except ValueError:
            continue

    return path.join(
        directory,
        "{}{}_{}_{}{}{}".format(prefix, datestr, number, rand, slug, extension),
    )


def create_with_editor(config, directory, migration_source, extension):
    editor = utils.get_editor(config)
    tmpfile = NamedTemporaryFile(
        mode="w",
        encoding="UTF-8",
        dir=directory,
        prefix=tempfile_prefix,
        suffix=extension,
        delete=False,
    )
    try:
        with tmpfile as f:
            f.write(migration_source)

        editor = [part.format(tmpfile.name) for part in shlex.split(editor)]
        if not any(tmpfile.name in part for part in editor):
            editor.append(tmpfile.name)

        mtime = stat(tmpfile.name).st_mtime
        sys.path.insert(0, directory)
        while True:
            try:
                subprocess.call(editor)
            except OSError:
                print("Error: could not open editor!")
            else:
                if stat(tmpfile.name).st_mtime == mtime:
                    print("abort: no changes made")
                    return None

            try:
                migration = Migration(None, tmpfile.name, None)
                migration.load()
                message = getattr(migration.module, "__doc__", "")
                break
            except Exception:
                message = ""
                print("Error loading migration")
                print(traceback.format_exc())
                print()
                r = utils.prompt("Retry editing?", "Ynq?")
                if r == "q":
                    return None
                elif r == "y":
                    continue
                elif r == "n":
                    break
                elif r == "?":
                    print("")
                    print("y: reopen the migration file in your editor")
                    print("n: save the migration as-is, without re-editing")
                    print("q: quit without saving the migration")
                    print("")
                    print("?: show this help")
                    continue

        sys.path = sys.path[1:]

        filename = make_filename(config, directory, message, extension)
        rename(tmpfile.name, filename)
        return filename
    finally:
        try:
            unlink(tmpfile.name)
        except OSError:
            pass



from yoyo import utils


class TestChangeParamStyle:
    def test_changes_to_qmark(self):
        sql = "SELECT :a, :b, :a"
        assert utils.change_param_style("qmark", sql, {"a": 1, "b": 2}) == (
            "SELECT ?, ?, ?",
            (1, 2, 1),
        )

    def test_changes_to_numeric(self):
        sql = "SELECT :a, :b, :a"
        assert utils.change_param_style("numeric", sql, {"a": 1, "b": 2}) == (
            "SELECT :1, :2, :3",
            (1, 2, 1),
        )

    def test_changes_to_format(self):
        sql = "SELECT :a, :b, :a"
        assert utils.change_param_style("format", sql, {"a": 1, "b": 2}) == (
            "SELECT %s, %s, %s",
            (1, 2, 1),
        )

    def test_changes_to_pyformat(self):
        sql = "SELECT :a, :b, :a"
        assert utils.change_param_style("pyformat", sql, {"a": 1, "b": 2}) == (
            "SELECT %(a)s, %(b)s, %(a)s",
            {"a": 1, "b": 2},
        )

    def test_changes_to_named(self):
        sql = "SELECT :a, :b, :a"
        assert utils.change_param_style("named", sql, {"a": 1, "b": 2}) == (
            "SELECT :a, :b, :a",
            {"a": 1, "b": 2},
        )



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import sys
import re
import warnings

import tabulate

from yoyo import (
    read_migrations,
    default_migration_table,
    ancestors,
    descendants,
)
from yoyo.migrations import topological_sort
from yoyo.scripts.main import InvalidArgument, get_backend
from yoyo import utils


def install_argparsers(global_parser, subparsers):
    # Standard options
    standard_options_parser = argparse.ArgumentParser(add_help=False)
    standard_options_parser.add_argument(
        "sources", nargs="*", help="Source directory of migration scripts"
    )

    standard_options_parser.add_argument(
        "-d",
        "--database",
        default=None,
        help="Database, eg 'sqlite:///path/to/sqlite.db' "
        "or 'postgresql://user@host/db'",
    )
    standard_options_parser.add_argument(
        "-p",
        "--prompt-password",
        dest="prompt_password",
        action="store_true",
        help="Prompt for the database password",
    )
    standard_options_parser.add_argument(
        "--migration-table",
        dest="migration_table",
        action="store",
        default=default_migration_table,
        help="Name of table to use for storing " "migration metadata",
    )

    # Options related to filtering the list of migrations
    filter_parser = argparse.ArgumentParser(add_help=False)
    filter_parser.add_argument(
        "-m",
        "--match",
        help="Select migrations matching PATTERN (regular expression)",
        metavar="PATTERN",
    )
    filter_parser.add_argument(
        "-r",
        "--revision",
        help=(
            "Apply/rollback migration with id REVISION and all its "
            "dependencies"
        ),
        metavar="REVISION",
    )

    # Options related to applying/rolling back migrations
    apply_parser = argparse.ArgumentParser(
        add_help=False, parents=[filter_parser]
    )
    apply_parser.add_argument(
        "-a",
        "--all",
        dest="all",
        action="store_true",
        help="Select all migrations, regardless of whether "
        "they have been previously applied",
    )
    apply_parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        help="Force apply/rollback of steps even if "
        "previous steps have failed",
    )

    parser_apply = subparsers.add_parser(
        "apply",
        help="Apply migrations",
        parents=[global_parser, standard_options_parser, apply_parser],
    )
    parser_apply.set_defaults(func=apply, command_name="apply")
    parser_apply.add_argument(
        "-1",
        "--one",
        help="Apply a single migration. "
        "If there are no unapplied migrations, reapply the last migration",
    )

    parser_develop = subparsers.add_parser(
        "develop",
        help="Apply migrations without prompting. "
        "If there are no unapplied migrations, reapply the last migration",
        parents=[global_parser, standard_options_parser, apply_parser],
    )
    parser_develop.set_defaults(func=develop, command_name="develop")
    parser_develop.add_argument(
        "-n", type=int, help="Act on the last N migrations", default=1,
    )

    parser_list = subparsers.add_parser(
        "list",
        help="List all available and applied migrations. "
        "Each listed migration is prefixed with either A (applied) or U "
        "(unapplied)",
        parents=[global_parser, standard_options_parser, filter_parser],
    )
    parser_list.set_defaults(func=list_migrations, command_name="list")

    parser_rollback = subparsers.add_parser(
        "rollback",
        parents=[global_parser, standard_options_parser, apply_parser],
        help="Rollback migrations",
    )
    parser_rollback.set_defaults(func=rollback, command_name="rollback")

    parser_reapply = subparsers.add_parser(
        "reapply",
        parents=[global_parser, standard_options_parser, apply_parser],
        help="Rollback then reapply migrations",
    )
    parser_reapply.set_defaults(func=reapply, command_name="reapply")

    parser_mark = subparsers.add_parser(
        "mark",
        parents=[global_parser, standard_options_parser, apply_parser],
        help="Mark migrations as applied, without running them",
    )
    parser_mark.set_defaults(func=mark, command_name="mark")

    parser_unmark = subparsers.add_parser(
        "unmark",
        parents=[global_parser, standard_options_parser, apply_parser],
        help="Unmark applied migrations, without rolling them back",
    )
    parser_unmark.set_defaults(func=unmark, command_name="unmark")

    parser_break_lock = subparsers.add_parser(
        "break-lock",
        parents=[global_parser, standard_options_parser],
        help="Break migration locks",
    )
    parser_break_lock.set_defaults(func=break_lock, command_name="break-lock")


def filter_migrations(migrations, pattern):

    if not pattern:
        return migrations

    search = re.compile(pattern).search
    return migrations.filter(lambda m: search(m.id) is not None)


def migrations_to_revision(migrations, revision, direction):
    if not revision:
        return migrations
    assert direction in {"apply", "rollback"}

    targets = [m for m in migrations if revision in m.id]
    if len(targets) == 0:
        raise InvalidArgument(
            "'{}' doesn't match any revisions.".format(revision)
        )
    if len(targets) > 1:
        raise InvalidArgument(
            "'{}' matches multiple revisions. "
            "Please specify one of {}.".format(
                revision, ", ".join(m.id for m in targets)
            )
        )

    target = targets[0]

    # apply: apply target and all its dependencies
    if direction == "apply":
        deps = ancestors(target, migrations)
        target_plus_deps = deps | {target}
        migrations = migrations.filter(lambda m: m in target_plus_deps)

    # rollback/reapply: rollback target and everything that depends on it
    else:
        deps = descendants(target, migrations)
        target_plus_deps = deps | {target}
        migrations = migrations.filter(lambda m: m in target_plus_deps)

    return migrations


def get_migrations(args, backend, direction=None):

    sources = args.sources
    dburi = args.database

    if not sources:
        raise InvalidArgument("Please specify the migration source directory")

    if not direction:
        direction = "apply" if args.func in {mark, apply} else "rollback"
    migrations = read_migrations(*sources)
    migrations = filter_migrations(migrations, args.match)
    migrations = migrations_to_revision(migrations, args.revision, direction)

    if direction == "apply":
        migrations = backend.to_apply(migrations)
    else:
        migrations = backend.to_rollback(migrations)

    if not args.batch_mode and not args.revision:
        migrations = prompt_migrations(backend, migrations, args.command_name)

    if (
        args.batch_mode
        and not args.revision
        and not args.all
        and args.func is rollback
    ):
        if len(migrations) > 1:
            warnings.warn(
                "Only rolling back a single migration."
                "To roll back multiple migrations, "
                "either use interactive mode or use "
                "--revision or --all"
            )
            migrations = migrations[:1]

    if not args.batch_mode and migrations:
        print("")
        print(
            "Selected",
            utils.plural(len(migrations), "%d migration:", "%d migrations:"),
        )
        for m in migrations:
            print("  [{m.id}]".format(m=m))
        prompt = "{} {} to {}".format(
            args.command_name.title(),
            utils.plural(
                len(migrations), "this migration", "these %d migrations"
            ),
            dburi,
        )
        if not utils.confirm(prompt, default="y"):
            return migrations.replace([])
    return migrations


def apply(args, config):
    backend = get_backend(args, config)
    with backend.lock():
        migrations = get_migrations(args, backend)
        backend.apply_migrations(migrations, args.force)


def reapply(args, config):
    backend = get_backend(args, config)
    with backend.lock():
        migrations = get_migrations(args, backend)
        backend.rollback_migrations(migrations, args.force)
        migrations = backend.to_apply(migrations)
        backend.apply_migrations(migrations, args.force)


def rollback(args, config):
    backend = get_backend(args, config)
    with backend.lock():
        migrations = get_migrations(args, backend)
        backend.rollback_migrations(migrations, args.force)


def develop(args, config):
    args.batch_mode = True
    backend = get_backend(args, config)
    with backend.lock():
        migrations = get_migrations(args, backend, "apply")
        if migrations:
            print(
                "Applying",
                utils.plural(len(migrations), "%d migration:", "%d migrations:"),
            )
            for m in migrations:
                print(f"  [{m.id}]")
            backend.apply_migrations(migrations, args.force)
        else:
            migrations = get_migrations(args, backend, "rollback")[: args.n]
            print(
                "Reapplying",
                utils.plural(len(migrations), "%d migration:", "%d migrations:"),
            )
            for m in migrations:
                print(f"  [{m.id}]")
            backend.rollback_migrations(migrations, args.force)
            migrations = get_migrations(args, backend, "apply")
            backend.apply_migrations(migrations, args.force)


def mark(args, config):
    backend = get_backend(args, config)
    with backend.lock():
        migrations = get_migrations(args, backend)
        backend.mark_migrations(migrations)


def unmark(args, config):
    backend = get_backend(args, config)
    with backend.lock():
        migrations = get_migrations(args, backend)
        backend.unmark_migrations(migrations)


def break_lock(args, config):
    backend = get_backend(args, config)
    backend.break_lock()


def list_migrations(args, config):
    backend = get_backend(args, config)
    migrations = read_migrations(*args.sources)
    migrations = filter_migrations(migrations, args.match)

    APPLIED, UNAPPLIED = "A", "U"
    if sys.stdout.isatty():
        APPLIED = "\033[92m{APPLIED}\033[0m".format(APPLIED=APPLIED)
        UNAPPLIED = "\033[91m{UNAPPLIED}\033[0m".format(UNAPPLIED=UNAPPLIED)

    with backend.lock():
        migrations = migrations.__class__(topological_sort(migrations))
        applied = backend.to_rollback(migrations)

        data = (
            (APPLIED if m in applied else UNAPPLIED, m.id, m.source_dir)
            for m in migrations
        )
        print(tabulate.tabulate(data, headers=("STATUS", "ID", "SOURCE")))


def prompt_migrations(backend, migrations, direction):
    """
    Iterate through the list of migrations and prompt the user to
    apply/rollback each. Return a list of user selected migrations.

    direction
        one of 'apply' or 'rollback'
    """

    class prompted_migration(object):
        def __init__(self, migration, default=None):
            super(prompted_migration, self).__init__()
            self.migration = migration
            self.choice = default

    to_prompt = [prompted_migration(m) for m in migrations]

    position = 0
    while position < len(to_prompt):
        mig = to_prompt[position]

        choice = mig.choice
        if choice is None:
            is_applied = backend.is_applied(mig.migration)
            if direction == "apply":
                choice = "n" if is_applied else "y"
            else:
                choice = "y" if is_applied else "n"
        options = "".join(
            o.upper() if o == choice else o.lower() for o in "ynvdaqjk?"
        )

        print("")
        print("[%s]" % (mig.migration.id,))
        response = utils.prompt(
            "Shall I %s this migration?" % (direction,), options
        )

        if response == "?":
            print("")
            print("y: %s this migration" % (direction,))
            print("n: don't %s it" % (direction,))
            print("")
            print("v: view this migration in full")
            print("")
            print(
                "d: %s the selected migrations, skipping any remaining"
                % (direction,)
            )
            print("a: %s all the remaining migrations" % (direction,))
            print("q: cancel without making any changes")
            print("")
            print("j: skip to next migration")
            print("k: back up to previous migration")
            print("")
            print("?: show this help")
            continue

        if response in "yn":
            mig.choice = response
            position += 1
            continue

        if response == "v":
            print(mig.migration.source)
            continue

        if response == "j":
            position = min(len(to_prompt), position + 1)
            continue

        if response == "k":
            position = max(0, position - 1)

        if response == "d":
            break

        if response == "a":
            for mig in to_prompt[position:]:
                mig.choice = "y"
            break

        if response == "q":
            for mig in to_prompt:
                mig.choice = "n"
            break

    return migrations.replace(m.migration for m in to_prompt if m.choice == "y")



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import chain
from tempfile import mkdtemp
from textwrap import dedent
from shutil import rmtree
import contextlib
import os.path

from yoyo.config import get_configparser
from yoyo.connections import get_backend

dburi = "sqlite:///:memory:"

config_file = os.path.join(
    os.path.dirname(__file__), *("../../test_databases.ini".split("/"))
)
config = get_configparser()
config.read([config_file])


def get_test_dburis(only=frozenset(), exclude=frozenset()):
    return [
        dburi
        for name, dburi in config.items("DEFAULT")
        if (only and name in only) or (not only and name not in exclude)
    ]


def get_test_backends(only=frozenset(), exclude=frozenset()):
    return [get_backend(dburi) for dburi in get_test_dburis(only, exclude)]


def clear_database(backend):
    with backend.transaction():
        for table in backend.list_tables():
            backend.execute("DROP TABLE {}".format(table))


class MigrationsContextManager(object):
    """
    Decorator/contextmanager taking a list of migrations.
    Creates a temporary directory writes
    each migration to a file (named '0.py', '1.py', '2.py' etc), calls the
    decorated function with the directory name as the first argument, and
    cleans up the temporary directory on exit.
    """

    def __init__(self, *migrations, **kwmigrations):
        self.migrations = migrations
        self.kwmigrations = kwmigrations

    def add_migration(self, id, code):
        _, extension = os.path.splitext(id)
        if extension in {".py", ".sql"}:
            filename = id
        else:
            filename = id + ".py"
        path = os.path.join(self.tmpdir, filename)
        with open(path, "w", encoding="UTF-8") as f:
            f.write(dedent(code).strip())

    def __enter__(self):
        tmpdir = self.tmpdir = mkdtemp()
        for mig_id, code in chain(
            enumerate(self.migrations), self.kwmigrations.items()
        ):
            self.add_migration(str(mig_id), code)
        return tmpdir

    def __exit__(self, *exc_info):
        rmtree(self.tmpdir)

    def __call__(self, func):
        def decorator(*args, **kwargs):
            with self:
                return func(*(args + (self.tmpdir,)), **kwargs)

        return decorator


@contextlib.contextmanager
def tempdir():
    tmpdir = mkdtemp()
    try:
        yield tmpdir
    finally:
        rmtree(tmpdir)


with_migrations = MigrationsContextManager
migrations_dir = MigrationsContextManager



#!/usr/bin/env python
# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from setuptools import setup

VERSIONFILE = "yoyo/__init__.py"


def get_version():
    with open(VERSIONFILE, 'rb') as f:
        return re.search("^__version__\s*=\s*['\"]([^'\"]*)['\"]",
                           f.read().decode('UTF-8'), re.M).group(1)


setup(name='yoyo-migrations', version=get_version())



"""
Version 2 schema.

Compatible with yoyo-migrations >=  6.0
"""
from datetime import datetime
from yoyo.migrations import get_migration_hash


def upgrade(backend):
    create_log_table(backend)
    create_version_table(backend)
    cursor = backend.execute(
        "SELECT id, ctime FROM {}".format(backend.migration_table_quoted)
    )
    migration_id = ""  # type: str
    created_at = datetime(1970, 1, 1)  # type: datetime
    for migration_id, created_at in iter(cursor.fetchone, None):  # type: ignore
        migration_hash = get_migration_hash(migration_id)
        log_data = dict(
            backend.get_log_data(),
            operation="apply",
            comment=(
                "this log entry created automatically by an "
                "internal schema upgrade"
            ),
            created_at_utc=created_at,
            migration_hash=migration_hash,
            migration_id=migration_id,
        )
        backend.execute(
            "INSERT INTO {0.log_table_quoted} "
            "(id, migration_hash, migration_id, operation, created_at_utc, "
            "username, hostname, comment) "
            "VALUES "
            "(:id, :migration_hash, :migration_id, 'apply', :created_at_utc, "
            ":username, :hostname, :comment)".format(backend),
            log_data,
        )

    backend.execute("DROP TABLE {0.migration_table_quoted}".format(backend))
    create_migration_table(backend)
    backend.execute(
        "INSERT INTO {0.migration_table_quoted} "
        "SELECT migration_hash, migration_id, created_at_utc "
        "FROM {0.log_table_quoted}".format(backend)
    )


def create_migration_table(backend):
    backend.execute(
        "CREATE TABLE {0.migration_table_quoted} ( "
        # sha256 hash of the migration id
        "migration_hash VARCHAR(64), "
        # The migration id (ie path basename without extension)
        "migration_id VARCHAR(255), "
        # When this id was applied
        "applied_at_utc TIMESTAMP, "
        "PRIMARY KEY (migration_hash))".format(backend)
    )


def create_log_table(backend):
    backend.execute(
        "CREATE TABLE {0.log_table_quoted} ( "
        "id VARCHAR(36), "
        "migration_hash VARCHAR(64), "
        "migration_id VARCHAR(255), "
        "operation VARCHAR(10), "
        "username VARCHAR(255), "
        "hostname VARCHAR(255), "
        "comment VARCHAR(255), "
        "created_at_utc TIMESTAMP, "
        "PRIMARY KEY (id))".format(backend)
    )


def create_version_table(backend):
    backend.execute(
        "CREATE TABLE {0.version_table_quoted} ("
        "version INT NOT NULL PRIMARY KEY, "
        "installed_at_utc TIMESTAMP)".format(backend)
    )
